import torch
import torch.nn as nn
import torch.nn.functional as F
from clipzyme.utils.registry import get_object, register_object
from clipzyme.utils.classes import set_nox_type
from clipzyme.utils.wln_processing import (
    generate_candidates_from_scores,
    get_batch_candidate_bonds,
    robust_edit_mol,
)
from clipzyme.models.abstract import AbstractModel
from torch_scatter import scatter_add
from torch_geometric.utils import to_dense_batch, to_dense_adj
from clipzyme.models.gat import GAT
from clipzyme.models.chemprop import WLNEncoder, DMPNNEncoder
from rdkit import Chem
import copy
import os
from rich import print as rprint


class WLDN_Cache:
    def __init__(self, path, extension="pt"):
        if not os.path.exists(path):
            os.makedirs(path)
        self.cache_dir = path
        self.files_extension = extension

    def _file_path(self, sample_id):
        return os.path.join(self.cache_dir, f"{sample_id}_candidates.pt")

    def exists(self, sample_id):
        return os.path.isfile(self._file_path(sample_id))

    def get(self, sample_id):
        return torch.load(self._file_path(sample_id))

    def add(self, sample_id, graph):
        torch.save(graph, self._file_path(sample_id))


@register_object("cheap_global_attn", "model")
class CheapGlobalAttention(AbstractModel):
    def __init__(self, args):
        super(CheapGlobalAttention, self).__init__()
        self.linear = nn.Linear(args.gat_hidden_dim, 1)

    def forward(self, node_feats, batch_index):
        # node_feats is (N, in_dim)
        scores = self.linear(node_feats)  # (N, 1)
        scores = torch.softmax(scores, dim=0)  # softmax over all nodes
        scores = scores.squeeze(1)  # (N, )
        out = scatter_add(node_feats * scores.unsqueeze(-1), batch_index, dim=0)
        return out


@register_object("pairwise_global_attn", "model")
class PairwiseAttention(AbstractModel):
    def __init__(self, args):
        super(PairwiseAttention, self).__init__()
        self.P_a = nn.Linear(args.gat_hidden_dim, args.gat_hidden_dim)
        self.P_b = nn.Linear(args.gat_complete_edge_dim, args.gat_hidden_dim)
        self.U = nn.Linear(args.gat_hidden_dim, 1)

    def forward(self, node_feats, graph):
        """Compute node contexts with global attention

        node_feats: N x F, where N is number of nodes, F is feature dimension
        graph: batched graph with edges of complete graph
        """
        # Batch index: N, mapping each node to corresponding graph index
        edge_index_complete = graph.edge_index_complete
        edge_attr_complete = graph.edge_attr_complete
        batch_index = graph.batch

        # Project features
        node_feats_transformed = self.P_a(node_feats)  # N x F
        edge_feats_complete = self.P_b(edge_attr_complete.float())  # E x F

        # convert to dense adj: E x F -> N x N x F
        dense_edge_attr = to_dense_adj(
            edge_index=edge_index_complete, edge_attr=edge_feats_complete
        ).squeeze(0)

        # node_features: sparse batch: N x D
        pairwise_node_feats = (
            node_feats_transformed.unsqueeze(1) + node_feats_transformed
        )  # N x N x F

        # Compute attention scores
        scores = torch.sigmoid(
            self.U(F.relu(pairwise_node_feats + dense_edge_attr))
        ).squeeze(-1)

        # Mask attention scores to prevent attention to nodes in different graphs
        mask = batch_index[:, None] != batch_index[None, :]  # N x N
        weights = scores.masked_fill(mask, 0)

        # Apply attention weights
        weighted_feats = torch.matmul(weights, node_feats)  # N x F

        return weighted_feats  # node_contexts


@register_object("gatv2_globalattn", "model")
class GATWithGlobalAttn(GAT):
    def __init__(self, args):
        super().__init__(args)
        self.global_attention = get_object(args.attn_type, "model")(args)

    def forward(self, graph):
        output = super().forward(graph)  # Graph NN (GAT) Local Network

        weighted_node_feats = self.global_attention(
            output["node_features"], graph
        )  # EQN 6

        output["node_features_attn"] = weighted_node_feats
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        GAT.add_args(parser)
        parser.add_argument(
            "--attn_type",
            type=str,
            action=set_nox_type("model"),
            default="pairwise_global_attn",
            help="type of global attention to use",
        )


@register_object("reaction_center_net", "model")
class ReactivityCenterNet(AbstractModel):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.gat_global_attention = get_object(args.gat_type, "model")(
            args
        )  # GATWithGlobalAttn(args)
        self.M_a = nn.Linear(args.gat_hidden_dim, args.gat_hidden_dim)
        self.M_b = nn.Linear(args.gat_complete_edge_dim, args.gat_hidden_dim)
        self.lin = nn.Linear(2 * args.gat_hidden_dim, args.gat_hidden_dim)
        self.U = nn.Sequential(
            nn.ReLU(), nn.Linear(args.gat_hidden_dim, args.num_predicted_bond_types)
        )

    def forward(self, batch):
        gat_output = self.gat_global_attention(
            batch["reactants"]
        )  # GAT + Global Attention over node features
        cs = gat_output["node_features"]
        c_tildes = gat_output["node_features_attn"]  # node contexts
        c_final = self.lin(
            torch.cat([cs, c_tildes], dim=-1)
        )  # N x 2*hidden_dim -> N x hidden_dim

        s_uv = self.forward_helper(
            c_final,
            batch["reactants"]["edge_index_complete"],
            batch["reactants"]["edge_attr_complete"],
            batch["reactants"]["batch"],
        )

        # precompute for top k metric
        candidate_bond_changes = get_batch_candidate_bonds(
            batch["reaction"], s_uv.detach(), batch["reactants"].batch
        )
        # make bonds that are "4" -> "1.5"
        for i in range(len(candidate_bond_changes)):
            candidate_bond_changes[i] = [
                (elem[0], elem[1], 1.5, elem[3]) if elem[2] == 4 else elem
                for elem in candidate_bond_changes[i]
            ]

        batch_real_bond_changes = []
        for i in range(len(batch["reactants"]["bond_changes"])):
            reaction_real_bond_changes = []
            for elem in batch["reactants"]["bond_changes"][i]:
                reaction_real_bond_changes.append(tuple(elem))
            batch_real_bond_changes.append(reaction_real_bond_changes)

        assert len(candidate_bond_changes) == len(batch_real_bond_changes)

        return {
            "s_uv": s_uv,
            "candidate_bond_changes": candidate_bond_changes,
            "real_bond_changes": batch_real_bond_changes,
            "c_final": c_final,
        }

    def forward_helper(self, node_features, edge_indices, edge_attr, batch_indices):
        # GAT with global attention
        node_features = self.M_a(node_features)  # N x hidden_dim -> N x hidden_dim
        edge_attr = self.M_b(edge_attr.float())  # E x 5 -> E x hidden_dim

        # convert to dense adj: E x hidden_dim -> N x N x hidden_dim
        dense_edge_attr = to_dense_adj(
            edge_index=edge_indices, edge_attr=edge_attr
        ).squeeze(0)

        # node_features: sparse batch: N x D
        pairwise_node_feats = node_features.unsqueeze(1) + node_features  # N x N x D
        # edge_attr: bond features: N x N x D
        s = self.U(dense_edge_attr + pairwise_node_feats).squeeze(-1)  # N x N
        # removed this line since the sizes become inconsistent later
        # s, mask = to_dense_batch(s, batch_indices) # B x max_batch_N x N x num_predicted_bond_types

        # make symmetric
        s = (s + s.transpose(0, 1)) / 2
        return s

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        AbstractModel.add_args(parser)
        parser.add_argument(
            "--num_predicted_bond_types",
            type=int,
            default=5,
            help="number of bond types to predict, this is t in the paper",
        )
        parser.add_argument(
            "--gat_type",
            type=str,
            action=set_nox_type("model"),
            default="gatv2_globalattn",
            help="Type of gat to use, mainly to init args",
        )
        parser.add_argument(
            "--topk_bonds",
            nargs="+",
            type=int,
            default=[1, 3, 5],
            help="topk bonds to consider for accuracy metric",
        )
        parser.add_argument(
            "--gat_complete_edge_dim",
            type=int,
            default=5,
            help="dimension of edges in complete graph",
        )


@register_object("wldn", "model")
class WLDN(AbstractModel):
    def __init__(self, args):
        super().__init__()
        self.args = args
        try:
            state_dict = torch.load(args.reactivity_model_path)
            self.reactivity_net = get_object(args.reactivity_net_type, "model")(
                state_dict["hyper_parameters"]["args"]
            )
            self.reactivity_net.load_state_dict(
                {
                    k[len("model.") :]: v
                    for k, v in state_dict["state_dict"].items()
                    if k.startswith("model")
                }
            )
            self.reactivity_net.requires_grad_(False)
            rprint(
                f"[magenta]Loaded Reactivity Model from {args.reactivity_model_path}[/magenta]"
            )
        except:
            self.reactivity_net = get_object(args.reactivity_net_type, "model")(
                args
            ).requires_grad_(False)
            rprint("[magenta]WARNING: Could not load pretrained model[/magenta]")
        self.reactivity_net.eval()

        self.add_scores_features = getattr(args, "add_scores_features", True)

        wln_diff_args = copy.deepcopy(args)

        if args.use_wln_encoder:
            # WLNEncoder
            self.wln = WLNEncoder(args)  # WLN for mol representation
            wln_diff_args = copy.deepcopy(args)
            wln_diff_args.wln_enc_node_dim = args.wln_enc_hidden_dim
            wln_diff_args.wln_enc_num_layers = 1
            self.wln_diff = WLNEncoder(wln_diff_args)
            self.final_transform = nn.Sequential(
                nn.Linear(
                    args.wln_enc_hidden_dim + 1
                    if self.add_scores_features
                    else args.wln_enc_hidden_dim,
                    args.wln_enc_hidden_dim,
                ),
                nn.ReLU(),
                nn.Linear(args.wln_enc_hidden_dim, 1),
            )

        elif args.use_chemprop_encoder:
            self.wln = DMPNNEncoder(args)  # WLN for mol representation
            wln_diff_args = copy.deepcopy(args)
            if args.model_name == "wldn":
                wln_diff_args.chemprop_node_dim = args.chemprop_hidden_dim
            else:
                wln_diff_args.chemprop_edge_dim = args.chemprop_hidden_dim + 1
            wln_diff_args.chemprop_num_layers = 1
            self.wln_diff = DMPNNEncoder(wln_diff_args)
            self.final_transform = nn.Sequential(
                nn.Linear(
                    args.chemprop_hidden_dim + 1
                    if args.add_scores_to_edge_attr
                    else args.chemprop_hidden_dim,
                    args.chemprop_hidden_dim,
                ),
                # nn.BatchNorm1d(args.chemprop_hidden_dim),
                nn.ReLU(),
                nn.Linear(args.chemprop_hidden_dim, 1),
            )

        elif args.use_gat_encoder:
            # GAT
            self.wln = GAT(args)  # WLN for mol representation GAT(args)
            wln_diff_args = copy.deepcopy(args)
            wln_diff_args.gat_node_dim = args.gat_hidden_dim
            wln_diff_args.gat_num_layers = 1
            self.wln_diff = GAT(wln_diff_args)
            self.final_transform = nn.Sequential(
                nn.Linear(
                    args.gat_hidden_dim + 1
                    if self.add_scores_features
                    else args.gat_hidden_dim,
                    args.gat_hidden_dim,
                ),
                nn.ReLU(),
                nn.Linear(args.gat_hidden_dim, 1),
            )

        if getattr(args, "load_wln_cache_in_dataset", False):
            self.use_cache = (args.cache_path is not None) and (
                not args.load_wln_cache_in_dataset
            )
            if self.use_cache:
                self.cache = WLDN_Cache(os.path.join(args.cache_path), "pt")
        if self.args.test:
            assert not self.args.train

        if args.ranker_model_path:
            try:
                state_dict = torch.load(args.ranker_model_path)
                self.wln.load_state_dict(
                    {
                        k[len("model.wln.") :]: v
                        for k, v in state_dict["state_dict"].items()
                        if k.startswith("model.wln.")
                    }
                )
                self.wln_diff.load_state_dict(
                    {
                        k[len("model.wln_diff.") :]: v
                        for k, v in state_dict["state_dict"].items()
                        if k.startswith("model.wln_diff.")
                    }
                )
                self.final_transform.load_state_dict(
                    {
                        k[len("model.final_transform.") :]: v
                        for k, v in state_dict["state_dict"].items()
                        if k.startswith("model.final_transform.")
                    }
                )
                rprint(
                    f"[magenta]Loaded Ranker from {args.ranker_model_path}[/magenta]"
                )
            except:
                pass

    def predict(self, batch, product_candidates_list, candidate_scores):
        predictions = []
        for idx, (candidates, scores) in enumerate(
            zip(product_candidates_list, candidate_scores)
        ):
            smiles_predictions = []
            # sort according to ranker score
            scores_indices = torch.argsort(scores.view(-1), descending=True)
            valid_candidate_combos = [
                candidates.candidate_bond_change[i] for i in scores_indices
            ]
            reactant_mol = Chem.MolFromSmiles(batch["reactants"].smiles[idx])
            for edits in valid_candidate_combos:
                edits_no_scores = [(x, y, t) for x, y, t, s in edits]
                smiles = robust_edit_mol(reactant_mol, edits_no_scores)
                # todo: smiles = set(sanitize_smiles_molvs(smiles) for smiles in smiles)
                if len(smiles) != 0:
                    smiles_predictions.append(smiles)
                else:
                    try:
                        Chem.Kekulize(reactant_mol)
                        smiles = robust_edit_mol(reactant_mol, edits_no_scores)
                        # todo: smiles = set(sanitize_smiles_molvs(smiles) for smiles in smiles)
                        smiles_predictions.append(smiles)
                    except Exception as e:
                        smiles_predictions.append([])
            predictions.append(smiles_predictions)

        return {"preds": predictions}

    def forward(self, batch):
        product_candidates_list = self.get_product_candidate_list(
            batch, batch["sample_id"]
        )

        reactant_node_feats = self.wln(batch["reactants"])[
            "node_features"
        ]  # N x D, where N is all the nodes in the batch
        dense_reactant_node_feats, mask = to_dense_batch(
            reactant_node_feats, batch=batch["reactants"].batch
        )  # B x max_batch_N x D
        candidate_scores = []
        for idx, product_candidates in enumerate(product_candidates_list):
            # get node features for candidate products
            product_candidates = product_candidates.to(reactant_node_feats.device)
            candidate_node_feats = self.wln(product_candidates)["node_features"]
            dense_candidate_node_feats, mask = to_dense_batch(
                candidate_node_feats, batch=product_candidates.batch
            )  # B x num_nodes x D

            num_nodes = dense_candidate_node_feats.shape[1]

            # compute difference vectors and replace the node features of the product graph with them
            difference_vectors = dense_candidate_node_feats - dense_reactant_node_feats[
                idx
            ][:num_nodes].unsqueeze(0)

            # undensify
            total_nodes = dense_candidate_node_feats.shape[0] * num_nodes
            difference_vectors = difference_vectors.view(total_nodes, -1)
            product_candidates.x = difference_vectors

            # apply a separate WLN to the difference graph
            wln_diff_output = self.wln_diff(product_candidates)
            diff_node_feats = wln_diff_output["node_features"]

            # compute the score for each candidate product
            # to dense
            diff_node_feats, _ = to_dense_batch(
                diff_node_feats, product_candidates.batch
            )  # num_candidates x max_num_nodes x D
            graph_feats = torch.sum(diff_node_feats, dim=-2)
            if self.add_scores_features:
                core_scores = [
                    sum(c[-1] for c in cand_changes)
                    for cand_changes in product_candidates.candidate_bond_change
                ]
                core_scores = torch.tensor(
                    core_scores, device=graph_feats.device
                ).unsqueeze(-1)
                score_order = torch.argsort(core_scores, dim=0)
                graph_feats = torch.concat([graph_feats, score_order], dim=-1)
            score = self.final_transform(graph_feats)
            if self.args.add_core_score:
                core_scores = [
                    sum(c[-1] for c in cand_changes)
                    for cand_changes in product_candidates.candidate_bond_change
                ]
                core_scores = torch.tensor(
                    core_scores, device=graph_feats.device
                ).unsqueeze(-1)
                score = score + torch.log_softmax(
                    core_scores, dim=0
                )  # torch.tensor(core_scores, device = score.device).unsqueeze(-1)
            candidate_scores.append(score)  # K x 1

        # ! PREDICT SMILES
        if self.args.predict:
            return self.predict(batch, product_candidates_list, candidate_scores)

        # note: dgl implementation adds reactivity score
        output = {
            "candidate_logit": candidate_scores,
            "product_candidates_list": product_candidates_list,
            # "s_uv": reactivity_output["s_uv"], # for debugging purposes
        }
        return output

    # seperate function because stays the same for different forward methods
    def get_product_candidate_list(self, batch, sample_ids):
        """
        Args:
            batch : collated samples from dataloader
            sample_ids: list of sample ids
        """
        if "product_candidates" in batch:
            return batch["product_candidates"]

        mode = (
            "test" if not self.args.train else "train"
        )  # this is used to get candidates, using robust_edit_mol when in predict later for actual smiles generation
        if self.use_cache:
            if not all(self.cache.exists(sid) for sid in sample_ids):
                with torch.no_grad():
                    reactivity_output = self.reactivity_net(
                        batch
                    )  # s_uv: N x N x 5, 'candidate_bond_changes', 'real_bond_changes'
                    if getattr(self.args, "mask_suv_scores", False):
                        reactivity_output["s_uv"][
                            ~batch["reactants"].mask.bool()
                        ] = -torch.inf

                # get candidate products as graph structures
                # each element in this list is a batch of candidate products (where each batch represents one reaction)
                product_candidates_list = generate_candidates_from_scores(
                    reactivity_output, batch, self.args, mode
                )
                [
                    self.cache.add(sid, product_candidates)
                    for sid, product_candidates in zip(
                        sample_ids, product_candidates_list
                    )
                ]
            else:
                product_candidates_list = [self.cache.get(sid) for sid in sample_ids]
        else:
            # each element in this list is a batch of candidate products (where each batch represents one reaction)
            with torch.no_grad():
                reactivity_output = self.reactivity_net(
                    batch
                )  # s_uv: N x N x 5, 'candidate_bond_changes', 'real_bond_changes'
                if getattr(self.args, "mask_suv_scores", False):
                    reactivity_output["s_uv"][
                        ~batch["reactants"].mask.bool()
                    ] = -torch.inf

            product_candidates_list = generate_candidates_from_scores(
                reactivity_output, batch, self.args, mode
            )

        return product_candidates_list

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        AbstractModel.add_args(parser)
        DMPNNEncoder.add_args(parser)
        WLNEncoder.add_args(parser)
        parser.add_argument(
            "--num_candidate_bond_changes", type=int, default=20, help="Core size"
        )
        parser.add_argument(
            "--max_num_bond_changes", type=int, default=5, help="Combinations"
        )
        parser.add_argument(
            "--max_num_change_combos_per_reaction", type=int, default=500, help="cutoff"
        )
        parser.add_argument(
            "--reactivity_net_type",
            type=str,
            action=set_nox_type("model"),
            default="reaction_center_net",
            help="Type of reactivity net to use, mainly to init args",
        )
        parser.add_argument(
            "--reactivity_model_path",
            type=str,
            help="path to pretrained reaction center prediction model",
        )
        parser.add_argument(
            "--add_core_score",
            action="store_true",
            default=False,
            help="whether to add core score to ranking prediction",
        )
        #### Wln encoder
        parser.add_argument(
            "--use_wln_encoder",
            action="store_true",
            default=False,
            help="use wln implementation.",
        )
        parser.add_argument(
            "--use_chemprop_encoder",
            action="store_true",
            default=False,
            help="use wln implementation.",
        )
        parser.add_argument(
            "--use_gat_encoder",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--add_scores_features",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--mask_suv_scores",
            action="store_true",
            default=False,
            help="use mask to zero out scores for certain reactants.",
        )
        parser.add_argument(
            "--add_scores_to_edge_attr",
            action="store_true",
            default=False,
            help="use gat implementation.",
        )
        parser.add_argument(
            "--ranker_model_path",
            type=str,
            help="path to pretrained ranker model if loading each model separately",
        )



# type: ignore

import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

import rdkit

# data
import clipzyme.datasets.enzymemap
import clipzyme.datasets.enzyme_screening
import clipzyme.datasets.reaction

# lightning
import clipzyme.lightning.base

# optimizers
import clipzyme.learning.optimizers.basic

# scheduler
import clipzyme.learning.schedulers.basic
import clipzyme.learning.schedulers.warmup

# losses
import clipzyme.learning.losses.basic
import clipzyme.learning.losses.contrastive


# metrics
import clipzyme.learning.metrics.basic
import clipzyme.learning.metrics.representation

# callbacks
import clipzyme.callbacks.basic

# models
import clipzyme.models.classifier
import clipzyme.models.gat
import clipzyme.models.chemprop
import clipzyme.models.fair_esm
import clipzyme.models.egnn
import clipzyme.models.protmol
import clipzyme.models.wln
import clipzyme.models.seq2seq

# comet
import clipzyme.loggers.wandb
import clipzyme.loggers.tensorboard

from clipzyme.datasets.reaction import ReactionDataset
from clipzyme.lightning.clipzyme import CLIPZyme
from clipzyme.utils.registry import get_object

__all__ = ["CLIPZyme", "ReactionDataset", "get_object"]



from collections import OrderedDict
from clipzyme.utils.classes import Nox
from torchmetrics import Metric
from clipzyme.utils.registry import register_object
import torch
from typing import Dict


def get_sample_similarity(h1, h2):
    cov = torch.mm(h1, h2.T)
    self_mean = torch.mean(torch.diag(cov))
    other_mean = torch.mean(cov - torch.diag(torch.diag(cov)))
    return self_mean, other_mean


def get_feature_similarity(h):
    return torch.std(h, dim=0).mean()


@register_object("contrastive_stat", "metric")
class ContrastiveStatistics(Nox):
    def __init__(self, args) -> None:
        super().__init__()
        # keys_for_contrastive_views_as_tuples = []
        # for keypair in args.keys_for_contrastive_views:
        #     keys_for_contrastive_views_as_tuples.append(tuple(keypair.split("/")))

        # args.keys_for_contrastive_views = keys_for_contrastive_views_as_tuples
        # self.metric_keys = args.keys_for_contrastive_views
        # self.metric_keys = self.metric_keys()

    @property
    def metric_keys(self):
        return ["reaction_projection", "sequence_projection"]

    def __call__(self, logging_dict, args):
        stats_dict = OrderedDict()
        k1, k2 = "reaction_projection", "sequence_projection"
        h1 = torch.nn.functional.normalize(logging_dict[k1], dim=1)
        h2 = torch.nn.functional.normalize(logging_dict[k2], dim=1)
        (
            stats_dict["self_similarity"],
            stats_dict["other_similarity"],
        ) = get_sample_similarity(h1, h2)
        stats_dict["{}_feature_std".format(k1)] = get_feature_similarity(h1)
        stats_dict["{}_feature_std".format(k2)] = get_feature_similarity(h2)

        return stats_dict

    # @staticmethod
    # def add_args(parser) -> None:
    #     """Add class specific args

    #     Args:
    #         parser (argparse.ArgumentParser): argument parser
    #     """
    #     parser.add_argument(
    #         "--keys_for_contrastive_views",
    #         type=float,
    #         default=2.0,
    #         help="Exponent for alignment loss",
    #     )


@register_object("alignment_uniformity", "metric")
class AlignmentUniformityLoss(Nox):
    def __init__(self, args) -> None:
        super().__init__()
        """
        https://arxiv.org/abs/2005.10242 
        """

    @property
    def metric_keys(self):
        return ["reaction_projection", "sequence_projection"]

    def __call__(self, logging_dict, args):
        stats_dict = OrderedDict()

        z1 = logging_dict["reaction_projection"]
        z2 = logging_dict["sequence_projection"]

        # Normalize
        z1 = torch.nn.functional.normalize(z1, dim=1)
        z2 = torch.nn.functional.normalize(z2, dim=1)

        alignment_loss = (z1 - z2).norm(dim=1).pow(args.alignment_alpha).mean()
        z1_uniformity_loss = (
            (torch.pdist(z1, p=2).pow(2))
            .mul(-args.uniformity_temperature)
            .exp()
            .mean()
            .log()
        )
        z2_uniformity_loss = (
            (torch.pdist(z2, p=2).pow(2))
            .mul(-args.uniformity_temperature)
            .exp()
            .mean()
            .log()
        )
        uniformity_loss = (z1_uniformity_loss + z2_uniformity_loss) / 2

        stats_dict["alignment"] = alignment_loss
        stats_dict["uniformity"] = uniformity_loss

        return stats_dict

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--alignment_alpha",
            type=float,
            default=2.0,
            help="Exponent for alignment loss",
        )
        parser.add_argument(
            "--uniformity_temperature",
            type=float,
            default=2.0,
            help="Temperature for uniformity loss",
        )


@register_object("clip_classification", "metric")
class ClipClassification(Metric, Nox):
    def __init__(self, args) -> None:
        """
        Computes standard classification metrics

        Args:
            predictions_dict: dictionary obtained from computing loss and model outputs
                * should contain the keys ['probs', 'preds', 'golds']
            args: argparser Namespace

        Returns:
            stats_dict (dict): contains (where applicable) values for accuracy, confusion matrix, precision, recall, f1, precision-recall auc, roc auc

        Note:
            Binary: two labels
            Multiclass: more than two labels
            Multilabel: potentially more than one label per sample (independent classes)
        """
        super().__init__()
        self.add_state("num_correct", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    @property
    def metric_keys(self):
        return ["clip_preds", "clip_golds"]

    def update(self, predictions_dict, args) -> Dict:
        preds = predictions_dict["clip_preds"]  # B
        golds = predictions_dict["clip_golds"].int()  # B

        self.num_correct += (golds == preds).sum()
        self.total += len(golds)

    def compute(self) -> Dict:
        stats_dict = {
            "clip_accuracy": self.num_correct.float() / self.total,
        }
        return stats_dict


@register_object("clip_quantile", "metric")
class ClipQuantile(Metric, Nox):
    def __init__(self, args) -> None:
        """
        Computes the quantile rank of a score relative to other scores pairwise similarity matrix
        """
        super().__init__()
        self.add_state("quantile", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    @property
    def metric_keys(self):
        return ["clip_probs", "clip_golds"]

    def update(self, predictions_dict, args) -> Dict:
        # select diagonal
        # check fraction > diagonal
        probs = predictions_dict["clip_probs"]  # N x K
        indices = predictions_dict["clip_golds"]
        # generalize for ddp gathering: pos_score = torch.diagonal(probs).unsqueeze(-1)
        pos_score = probs.gather(1, indices.unsqueeze(1))
        num_less_than_score = probs < pos_score
        score_quantile = num_less_than_score.float().mean(1).sum()
        self.quantile += score_quantile
        self.total += probs.shape[0]

    def compute(self) -> Dict:
        stats_dict = {
            "clip_quantile": self.quantile.float() / self.total,
        }
        return stats_dict



import math
from typing import TypeVar, Optional, Iterator, Sequence

import torch
from torch.utils.data import Dataset
import torch.distributed as dist


T_co = TypeVar("T_co", covariant=True)


class DistributedWeightedSampler(torch.utils.data.distributed.DistributedSampler):
    r"""Extension of pytorch's native distributed sampler, but supports weighted sampling
    .. note::
        Dataset is assumed to be of constant size.
    Arguments:
        dataset: Dataset used for sampling.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, :attr:`rank` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`num_replicas`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
            indices.
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.
        drop_last (bool, optional): if ``True``, then the sampler will drop the
            tail of the data to make it evenly divisible across the number of
            replicas. If ``False``, the sampler will add extra indices to make
            the data evenly divisible across the replicas. Default: ``False``.
    .. warning::
        In distributed mode, calling the :meth:`set_epoch` method at
        the beginning of each epoch **before** creating the :class:`DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.
    Example::
        >>> sampler = DistributedSampler(dataset) if is_distributed else None
        >>> loader = DataLoader(dataset, shuffle=(sampler is None),
        ...                     sampler=sampler)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(
        self,
        dataset: Dataset,
        weights: Sequence[float],
        replacement: bool = True,
        generator=None,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        seed: int = 0,
        drop_last: bool = False,
    ) -> None:
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.drop_last = drop_last
        self.weights = torch.Tensor(weights)
        self.replacement = replacement
        self.generator = generator
        # If the dataset length is evenly divisible by # of replicas, then there
        # is no need to drop any data, since the dataset will be split equally.
        if self.drop_last and len(self.dataset) % self.num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            self.num_samples = math.ceil(
                (len(self.dataset) - self.num_replicas) / self.num_replicas
            )
        else:
            self.num_samples = math.ceil(len(self.dataset) / self.num_replicas)
        self.total_size = self.num_samples * self.num_replicas
        self.seed = seed

    def __iter__(self) -> Iterator[T_co]:
        indices = list(range(len(self.dataset)))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            indices += indices[: (self.total_size - len(indices))]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[: self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        weights = self.weights[self.rank : self.total_size : self.num_replicas]

        assert len(indices) == self.num_samples

        rand_tensor = torch.multinomial(
            self.weights, self.num_samples, self.replacement, generator=self.generator
        )
        return iter(rand_tensor)



"""
From ColabFold: https://github.com/sokrypton/ColabFold/tree/main
    colabfold/batch.py
    colabfold/utils/py
"""

from typing import List, Optional, Tuple, Union
from pathlib import Path
import pandas
import random


def safe_filename(file: str) -> str:
    return "".join([c if c.isalnum() or c in ["_", ".", "-"] else "_" for c in file])


def parse_fasta(fasta_string: str) -> Tuple[List[str], List[str]]:
    """Parses FASTA string and returns list of strings with amino-acid sequences.

    Arguments:
      fasta_string: The string contents of a FASTA file.

    Returns:
      A tuple of two lists:
      * A list of sequences.
      * A list of sequence descriptions taken from the comment lines. In the
        same order as the sequences.
    """
    sequences = []
    descriptions = []
    index = -1
    for line in fasta_string.splitlines():
        line = line.strip()
        if line.startswith("#"):
            continue
        if line.startswith(">"):
            index += 1
            descriptions.append(line[1:])  # Remove the '>' at the beginning.
            sequences.append("")
            continue
        elif not line:
            continue  # Skip blank lines.
        sequences[index] += line

    return sequences, descriptions


def get_queries(
    input_path: Union[str, Path], sort_queries_by: str = "length"
) -> Tuple[List[Tuple[str, str, Optional[List[str]]]], bool]:
    """Reads a directory of fasta files, a single fasta file or a csv file and returns a tuple
    of job name, sequence and the optional a3m lines"""

    input_path = Path(input_path)
    if not input_path.exists():
        raise OSError(f"{input_path} could not be found")

    if input_path.is_file():
        if input_path.suffix == ".csv" or input_path.suffix == ".tsv":
            sep = "\t" if input_path.suffix == ".tsv" else ","
            df = pandas.read_csv(input_path, sep=sep)
            assert "id" in df.columns and "sequence" in df.columns
            queries = [
                (seq_id, sequence.upper().split(":"), None)
                for seq_id, sequence in df[["id", "sequence"]].itertuples(index=False)
            ]
            for i in range(len(queries)):
                if len(queries[i][1]) == 1:
                    queries[i] = (queries[i][0], queries[i][1][0], None)
        elif input_path.suffix == ".a3m":
            (seqs, header) = parse_fasta(input_path.read_text())
            if len(seqs) == 0:
                raise ValueError(f"{input_path} is empty")
            query_sequence = seqs[0]
            # Use a list so we can easily extend this to multiple msas later
            a3m_lines = [input_path.read_text()]
            queries = [(input_path.stem, query_sequence, a3m_lines)]
        elif input_path.suffix in [".fasta", ".faa", ".fa"]:
            (sequences, headers) = parse_fasta(input_path.read_text())
            queries = []
            for sequence, header in zip(sequences, headers):
                sequence = sequence.upper()
                if sequence.count(":") == 0:
                    # Single sequence
                    queries.append((header, sequence, None))
                else:
                    # Complex mode
                    queries.append((header, sequence.upper().split(":"), None))
        else:
            raise ValueError(f"Unknown file format {input_path.suffix}")
    else:
        assert input_path.is_dir(), "Expected either an input file or a input directory"
        queries = []
        for file in sorted(input_path.iterdir()):
            if not file.is_file():
                continue
            if file.suffix.lower() not in [".a3m", ".fasta", ".faa"]:
                # print(f"non-fasta/a3m file in input directory: {file}")
                continue
            (seqs, header) = parse_fasta(file.read_text())
            if len(seqs) == 0:
                print(f"{file} is empty")
                continue
            query_sequence = seqs[0]
            if len(seqs) > 1 and file.suffix in [".fasta", ".faa", ".fa"]:
                print(
                    f"More than one sequence in {file}, ignoring all but the first sequence"
                )

            if file.suffix.lower() == ".a3m":
                a3m_lines = [file.read_text()]
                queries.append((file.stem, query_sequence.upper(), a3m_lines))
            else:
                if query_sequence.count(":") == 0:
                    # Single sequence
                    queries.append((file.stem, query_sequence, None))
                else:
                    # Complex mode
                    queries.append((file.stem, query_sequence.upper().split(":"), None))

    # sort by seq. len
    if sort_queries_by == "length":
        queries.sort(key=lambda t: len("".join(t[1])))

    elif sort_queries_by == "random":
        random.shuffle(queries)

    is_complex = False
    for job_number, (_, query_sequence, a3m_lines) in enumerate(queries):
        if isinstance(query_sequence, list):
            is_complex = True
            break
        if a3m_lines is not None and a3m_lines[0].startswith("#"):
            a3m_line = a3m_lines[0].splitlines()[0]
            tab_sep_entries = a3m_line[1:].split("\t")
            if len(tab_sep_entries) == 2:
                query_seq_len = tab_sep_entries[0].split(",")
                query_seq_len = list(map(int, query_seq_len))
                query_seqs_cardinality = tab_sep_entries[1].split(",")
                query_seqs_cardinality = list(map(int, query_seqs_cardinality))
                is_single_protein = (
                    True
                    if len(query_seq_len) == 1 and query_seqs_cardinality[0] == 1
                    else False
                )
                if not is_single_protein:
                    is_complex = True
                    break
    return queries, is_complex


def pad_sequences(
    a3m_lines: List[str], query_sequences: List[str], query_cardinality: List[int]
) -> str:
    _blank_seq = [
        ("-" * len(seq))
        for n, seq in enumerate(query_sequences)
        for _ in range(query_cardinality[n])
    ]
    a3m_lines_combined = []
    pos = 0
    for n, seq in enumerate(query_sequences):
        for j in range(0, query_cardinality[n]):
            lines = a3m_lines[n].split("\n")
            for a3m_line in lines:
                if len(a3m_line) == 0:
                    continue
                if a3m_line.startswith(">"):
                    a3m_lines_combined.append(a3m_line)
                else:
                    a3m_lines_combined.append(
                        "".join(_blank_seq[:pos] + [a3m_line] + _blank_seq[pos + 1 :])
                    )
            pos += 1
    return "\n".join(a3m_lines_combined)


def pair_sequences(
    a3m_lines: List[str], query_sequences: List[str], query_cardinality: List[int]
) -> str:
    a3m_line_paired = [""] * len(a3m_lines[0].splitlines())
    for n, seq in enumerate(query_sequences):
        lines = a3m_lines[n].splitlines()
        for i, line in enumerate(lines):
            if line.startswith(">"):
                if n != 0:
                    line = line.replace(">", "\t", 1)
                a3m_line_paired[i] = a3m_line_paired[i] + line
            else:
                a3m_line_paired[i] = a3m_line_paired[i] + line * query_cardinality[n]
    return "\n".join(a3m_line_paired)


def pair_msa(
    query_seqs_unique: List[str],
    query_seqs_cardinality: List[int],
    paired_msa: Optional[List[str]],
    unpaired_msa: Optional[List[str]],
) -> str:
    if paired_msa is None and unpaired_msa is not None:
        a3m_lines = pad_sequences(
            unpaired_msa, query_seqs_unique, query_seqs_cardinality
        )
    elif paired_msa is not None and unpaired_msa is not None:
        a3m_lines = (
            pair_sequences(paired_msa, query_seqs_unique, query_seqs_cardinality)
            + "\n"
            + pad_sequences(unpaired_msa, query_seqs_unique, query_seqs_cardinality)
        )
    elif paired_msa is not None and unpaired_msa is None:
        a3m_lines = pair_sequences(
            paired_msa, query_seqs_unique, query_seqs_cardinality
        )
    else:
        raise ValueError(f"Invalid pairing")
    return a3m_lines


def msa_to_str(
    unpaired_msa: List[str],
    paired_msa: List[str],
    query_seqs_unique: List[str],
    query_seqs_cardinality: List[int],
) -> str:
    msa = "#" + ",".join(map(str, map(len, query_seqs_unique))) + "\t"
    msa += ",".join(map(str, query_seqs_cardinality)) + "\n"
    # build msa with cardinality of 1, it makes it easier to parse and manipulate
    query_seqs_cardinality = [1 for _ in query_seqs_cardinality]
    msa += pair_msa(query_seqs_unique, query_seqs_cardinality, paired_msa, unpaired_msa)
    return msa



import torch
import torch.nn as nn
import copy
from typing import List
from clipzyme.models.abstract import AbstractModel
from clipzyme.utils.classes import set_nox_type
from clipzyme.utils.registry import register_object, get_object
from clipzyme.utils.amino_acids import AA_TO_SMILES
from torch_geometric.data import Data, HeteroData, Batch
from clipzyme.utils.pyg import from_smiles
from clipzyme.utils.smiles import get_rdkit_feature


@register_object("fair_esm", "model")
class FairEsm(AbstractModel):
    """
    Refer to https://github.com/facebookresearch/esm#available-models
    """

    def __init__(self, args):
        super(FairEsm, self).__init__()
        self.args = args
        torch.hub.set_dir(args.pretrained_hub_dir)
        self.model, self.alphabet = torch.hub.load(
            "facebookresearch/esm:v2.0.0", args.esm_name
        )
        self.batch_converter = self.alphabet.get_batch_converter()
        self.register_buffer("devicevar", torch.zeros(1, dtype=torch.int8))
        if args.freeze_esm:
            self.model.eval()

        self.repr_layer = args.esm_hidden_layer
        self.use_cls_token = args.use_esm_cls_token

    def forward(self, x):
        """
        x: list of str (protein sequences)
        """
        output = {}
        if isinstance(x, list):
            pass
        elif isinstance(x, dict):
            try:
                x = x["sequence"]
            except:
                raise ValueError(
                    "FairEsm forward received dict without 'sequence' key "
                )

        fair_x = self.truncate_protein(x)
        batch_labels, batch_strs, batch_tokens = self.batch_converter(fair_x)
        batch_tokens = batch_tokens.to(self.devicevar.device)

        if self.args.freeze_esm:
            self.model.requires_grad_(False)
            with torch.no_grad():
                result = self.model(
                    batch_tokens,
                    repr_layers=[self.repr_layer],
                    return_contacts=self.args.esm_return_contacts,
                )
        else:
            result = self.model(
                batch_tokens,
                repr_layers=[self.repr_layer],
                return_contacts=self.args.esm_return_contacts,
            )

        # Generate per-sequence representations via averaging
        if self.use_cls_token:
            output["hidden"] = result["representations"][self.repr_layer][0]
        else:
            # remove cls, eos, and padding embeddings
            sequence_mask = torch.ne(batch_tokens, self.alphabet.cls_idx).long()
            sequence_mask *= torch.ne(batch_tokens, self.alphabet.eos_idx).long()
            sequence_mask *= torch.ne(batch_tokens, self.alphabet.padding_idx).long()
            sequence_mask = sequence_mask.unsqueeze(-1)
            # remove cls and eos tokens
            output["hidden"] = (
                result["representations"][self.repr_layer] * sequence_mask
            ).sum(1) / sequence_mask.sum(1)
            output["mask_hiddens"] = sequence_mask

        output["tokens"] = batch_tokens
        output["token_hiddens"] = result["representations"][self.repr_layer]

        return output

    def truncate_protein(self, x, max_length=1024):
        # max length allowed is 1024
        return [
            (i, s[: 1024 - 2]) if not isinstance(x[0], list) else (i, s[0][: 1024 - 2])
            for i, s in enumerate(x)
        ]

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--pretrained_hub_dir",
            type=str,
            default="/home/snapshots/metabolomics",
            help="directory to torch hub where pretrained models are saved",
        )
        parser.add_argument(
            "--esm_name",
            type=str,
            default="esm2_t12_35M_UR50D",
            help="directory to torch hub where pretrained models are saved",
        )
        parser.add_argument(
            "--freeze_esm",
            action="store_true",
            default=False,
            help="do not update encoder weights",
        )
        parser.add_argument(
            "--esm_hidden_layer",
            type=int,
            default=12,
            help="do not update encoder weights",
        )
        parser.add_argument(
            "--use_esm_cls_token",
            action="store_true",
            default=False,
            help="use cls token as representation",
        )
        parser.add_argument(
            "--esm_return_contacts",
            action="store_true",
            default=False,
            help="return contacts",
        )


@register_object("fair_esm2", "model")
class FairEsm2(FairEsm):
    def truncate_protein(self, x, max_length=torch.inf):
        return [
            (i, s) if not isinstance(x[0], list) else (i, s[0]) for i, s in enumerate(x)
        ]


@register_object("protein_encoder", "model")
class ProteinEncoder(AbstractModel):
    def __init__(self, args):
        super(ProteinEncoder, self).__init__()
        self.args = args
        self.encoder = get_object(args.protein_encoder_type, "model")(args)
        cargs = copy.deepcopy(args)
        cargs.mlp_input_dim = args.protein_hidden_dim
        args.freeze_esm = args.freeze_encoder
        self.mlp = get_object(args.protein_classifer, "model")(cargs)
        if self.args.freeze_encoder:
            self.encoder.eval()

    def forward(self, batch):
        output = {}
        if self.args.freeze_encoder:
            self.encoder.requires_grad_(False)
            with torch.no_grad():
                output_esm = self.encoder(batch["x"])
        else:
            output_esm = self.encoder(batch["x"])
        # output["protein_hidden"] = output_esm["hidden"]
        output.update(self.mlp({"x": output_esm["hidden"]}))
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--protein_encoder_type",
            type=str,
            default="fair_esm",
            help="name of the protein encoder",
            action=set_nox_type("model"),
        )
        parser.add_argument(
            "--freeze_encoder",
            action="store_true",
            default=False,
            help="do not update encoder weights",
        )
        parser.add_argument(
            "--protein_hidden_dim",
            type=int,
            default=480,
            help="hidden dimension of the protein",
        )
        parser.add_argument(
            "--protein_classifer",
            type=str,
            default="mlp_classifier",
            help="name of classifier",
            action=set_nox_type("model"),
        )



AA_TO_SMILES = {
    "A": "C[C@@H](C(=O)O)N",  # Alanine
    "R": "C(C[C@@H](C(=O)O)N)CN=C(N)N",  # Arginine
    "N": "C([C@@H](C(=O)O)N)C(=O)N",  # Asparagine
    "D": "C([C@@H](C(=O)O)N)C(=O)O",  # Aspartic acid
    "B": "X",  # Aspartate or Asparagine
    "C": "C([C@@H](C(=O)O)N)S",  # Cysteine
    "E": "C(CC(=O)O)[C@@H](C(=O)O)N",  # Glutamic acid
    "Q": "C(CC(=O)N)[C@@H](C(=O)O)N",  # Glutamine
    "Z": "X",  # Glutamate or Glutamine
    "G": "C(C(=O)O)N",  # Glycine
    "H": "C([C@@H](C(=O)O)N)C(=O)N",  # Histidine
    "I": "CC[C@H](C)[C@@H](C(=O)O)N",  # Isoleucine
    "L": "CC(C)C[C@@H](C(=O)O)N",  # Leucine
    "K": "C(CCN)C[C@@H](C(=O)O)N",  # Lysine
    "M": "CSCC[C@@H](C(=O)O)N",  # Methionine
    "F": "C1=CC=C(C=C1)C[C@@H](C(=O)O)N",  # Phenylalanine
    "P": "C1C[C@H](NC1)C(=O)O",  # Proline
    "S": "C([C@@H](C(=O)O)N)O",  # Serine
    "T": "C[C@H]([C@@H](C(=O)O)N)O",  # Threonine
    "W": "C1=CC=C2C(=C1)C(=CN2)C[C@@H](C(=O)O)N",  # Tryptophan
    "Y": "C1=CC(=CC=C1C[C@@H](C(=O)O)N)O",  # Tyrosine
    "V": "CC(C)[C@@H](C(=O)O)N",  # Valine
    "X": "X",
    "U": "C([C@@H](C(=O)O)N)[SeH]",  # Selenocysteine
}

AA_TO_TRIPLET = {
    "A": "ala",  # Alanine
    "R": "arg",  # Arginine
    "N": "asn",  # Asparagine
    "D": "asp",  # Aspartic acid
    "B": ["asp", "asn"],  # Aspartate or Asparagine
    "C": "cys",  # Cysteine
    "E": "glu",  # Glutamic acid
    "Q": "gln",  # Glutamine
    "Z": ["glu", "gln"],  # Glutamate or Glutamine
    "G": "gly",  # Glycine
    "H": "his",  # Histidine
    "I": "ile",  # Isoleucine
    "L": "leu",  # Leucine
    "K": "lys",  # Lysine
    "M": "met",  # Methionine
    "F": "phe",  # Phenylalanine
    "P": "pro",  # Proline
    "S": "ser",  # Serine
    "T": "thr",  # Threonine
    "W": "trp",  # Tryptophan
    "Y": "tyr",  # Tyrosine
    "V": "val",  # Valine
    "X": "x",
    "U": "sec",  # Selenocysteine
}



import torch
import pytorch_lightning as pl
import torch.nn.functional as F
import numpy as np
from collections import OrderedDict
import pickle
import os
from clipzyme.utils.registry import get_object, register_object
from clipzyme.utils.classes import Nox, set_nox_type


@register_object("base", "lightning")
class Base(pl.LightningModule, Nox):
    """
    PyTorch Lightning module used as base for running training and test loops

    Args:
        args: argparser Namespace
    """

    def __init__(self, args):
        super(Base, self).__init__()
        self.training_step_outputs = []
        self.validation_step_outputs = []
        self.test_step_outputs = []
        self.save_hyperparameters()
        self.args = args
        self.model = get_object(args.model_name, "model")(args)

    def on_fit_start(self):
        self.setup_losses_and_metrics()

    def on_test_start(self):
        self.setup_losses_and_metrics()

    def on_predict_start(self):
        self.setup_losses_and_metrics()

    def setup_losses_and_metrics(self):
        # losses
        self.loss_fns = {
            "train": [get_object(l, "loss")() for l in self.args.loss_names]
        }
        self.loss_fns["val"] = self.loss_fns["train"]
        self.loss_fns["test"] = (
            self.loss_fns["train"]
            if self.args.loss_names_for_eval is None
            else [get_object(l, "loss")() for l in self.args.loss_names_for_eval]
        )

        # metrics
        self.metrics = {
            "train": [
                get_object(m, "metric")(self.args).to(self.device)
                for m in self.args.metric_names
            ]
        }
        self.metrics["val"] = [
            get_object(m, "metric")(self.args).to(self.device)
            for m in self.args.metric_names
        ]
        self.metrics["test"] = (
            [
                get_object(m, "metric")(self.args).to(self.device)
                for m in self.args.metric_names
            ]
            if self.args.metric_names_for_eval is None
            else [
                get_object(l, "metric")(self.args)
                for l in self.args.metric_names_for_eval
            ]
        )

        self.metric_keys = list(
            set(
                [
                    key
                    for split, metrics in self.metrics.items()
                    for metric in metrics
                    for key in metric.metric_keys
                ]
            )
        )

    @property
    def LOG_KEYS(self):
        return [
            "loss",
            "accuracy",
            "mean",
            "std",
            "precision",
            "recall",
            "f1",
            "auc",
            "similarity",
            "tau",
            "mse",
            "mae",
            "r2",
            "c_index",
            "hit",
            "pearson",
            "spearman",
            "cosine_similarity",
            "top",
            "coverage",
            "found",
            "quantile",
        ]

    @property
    def UNLOG_KEYS(self):
        default = ["activ", "hidden", "tokens", "token_hiddens", "mask_hiddens"]
        keys_to_unlog = []
        for k in default:
            if k not in self.metric_keys:
                keys_to_unlog.append(k)
        return keys_to_unlog

    def step(self, batch, batch_idx, optimizer_idx):
        """
        Defines a single training or validation step:
            Computes losses given batch and model outputs

        Returns:
            logged_output: dict with losses and predictions

        Args:
            batch: dict obtained from DataLoader. batch must contain they keys ['x', 'sample_id']
        """
        logged_output = OrderedDict()
        model_output = self.model(batch)
        loss, logging_dict, predictions_dict = self.compute_loss(model_output, batch)
        predictions_dict = self.store_in_predictions(predictions_dict, batch)
        predictions_dict = self.store_in_predictions(predictions_dict, model_output)
        self.call_metric(predictions_dict, "update")
        if self.args.compute_metrics_every_step:
            metrics = self.call_metric(predictions_dict, "compute")
            self.log_outputs(metrics, self.phase)
            self.log_outputs(logging_dict, self.phase)
            for metric_fn in self.metrics[self.phase]:
                metric_fn.reset()

        logged_output["loss"] = loss
        logged_output.update(logging_dict)

        if (
            (self.args.log_gen_image)
            and (self.trainer.is_global_zero)
            and (batch_idx == 0)
            and (self.current_epoch % 100 == 0)
        ):
            self.log_image(model_output, batch)

        if (self.args.log_predictions_to_table) and (
            self.global_step % self.args.log_predictions_to_table_every == 0
        ):
            table_columns = [
                "sample",
                "gold mol",
                "gold smiles",
                "pred",
                "step",
                "phase",
            ]
            data_types = ["str", "smiles", "str", "str", "str", "str"]
            table_data = [
                batch["sample_id"],
                batch["smiles"],
                batch["smiles"],
                model_output["pred_smiles"],
                [self.current_epoch] * len(batch["smiles"]),
                [self.phase] * len(batch["smiles"]),
            ]
            self.logger.log_table(
                table_columns,
                table_data,
                data_types,
                table_name=f"{self.phase}_samples",
            )

        return logged_output

    def forward(self, batch, batch_idx=0):
        """
        Forward defines the prediction/inference actions
            Similar to self.step() but also allows for saving predictions and hiddens
            Computes losses given batch and model outputs

        Returns:
            logged_output: dict with losses and predictions

        Args:
            batch: dict obtained from DataLoader. batch must contain they keys ['x', 'sample_id']
        """
        logged_output = OrderedDict()
        model_output = self.model(batch)
        if not self.args.predict:
            loss, logging_dict, predictions_dict = self.compute_loss(
                model_output, batch
            )
            predictions_dict = self.store_in_predictions(predictions_dict, batch)
            predictions_dict = self.store_in_predictions(predictions_dict, model_output)
            logged_output["loss"] = loss
        else:
            logging_dict = {}
            predictions_dict = {}
            predictions_dict = self.store_in_predictions(predictions_dict, batch)
            predictions_dict = self.store_in_predictions(predictions_dict, model_output)

        self.call_metric(predictions_dict, "update")

        logged_output.update(logging_dict)
        logged_output["preds_dict"] = predictions_dict
        if self.args.save_hiddens:
            logged_output["preds_dict"].update(model_output)

        if (self.args.log_gen_image) and (batch_idx == 0):
            self.log_image(model_output, batch)

        return logged_output

    def training_step(self, batch, batch_idx, optimizer_idx=None):
        """
        Single training step
        """
        self.phase = "train"
        output = self.step(batch, batch_idx, optimizer_idx)
        self.training_step_outputs.append(output)
        return output

    def validation_step(self, batch, batch_idx, optimizer_idx=None):
        """
        Single validation step
        """
        self.phase = "val"
        output = self.step(batch, batch_idx, optimizer_idx)
        self.validation_step_outputs.append(output)
        return output

    def test_step(self, batch, batch_idx):
        """
        Single testing step

        * save_predictions will save the dictionary output['preds_dict'], which typically includes sample_ids, probs, predictions, etc.
        * save_hiddens: will save the value of output['preds_dict']['hidden']
        """
        self.phase = "test"
        output = self.forward(batch, batch_idx)
        if self.args.save_predictions:
            self.save_predictions(output["preds_dict"])
        elif self.args.save_hiddens:
            self.save_hiddens(output["preds_dict"])
        output = {k: v for k, v in output.items() if k not in self.UNLOG_KEYS}
        output["preds_dict"] = {
            k: v for k, v in output["preds_dict"].items() if k not in self.UNLOG_KEYS
        }
        self.test_step_outputs.append(output)
        return output

    def on_train_epoch_end(self):
        """
        End of single training epoch
            - Aggregates predictions and losses from all steps
            - Computes the metric (auc, accuracy, etc.)
        """
        self.phase = "train"
        if len(self.training_step_outputs) == 0:
            return
        outputs = gather_step_outputs(self.training_step_outputs)
        outputs["loss"] = outputs["loss"].mean()
        outputs.update(self.call_metric({}, "compute"))
        self.log_outputs(outputs, "train")
        for metric_fn in self.metrics["train"]:
            metric_fn.reset()
        self.training_step_outputs.clear()
        return

    def on_validation_epoch_end(self):
        """
        End of single validation epoch
            - Aggregates predictions and losses from all steps
            - Computes the metric (auc, accuracy, etc.)
        """
        self.phase = "val"
        if len(self.validation_step_outputs) == 0:
            return
        outputs = gather_step_outputs(self.validation_step_outputs)
        outputs["loss"] = outputs["loss"].mean()
        outputs.update(self.call_metric({}, "compute"))
        self.log_outputs(outputs, "val")
        for metric_fn in self.metrics["val"]:
            metric_fn.reset()
        self.validation_step_outputs.clear()
        return

    def on_test_epoch_end(self):
        """
        End of testing
            - Aggregates predictions and losses from all batches
            - Computes the metric if defined in args
        """
        self.phase = "test"
        if len(self.test_step_outputs) == 0:
            return
        outputs = gather_step_outputs(self.test_step_outputs)
        if isinstance(outputs.get("loss", 0), torch.Tensor):
            outputs["loss"] = outputs["loss"].mean()
        if not self.args.predict:
            outputs.update(self.call_metric({}, "compute"))
        self.log_outputs(outputs, "test")
        for metric_fn in self.metrics["test"]:
            metric_fn.reset()
        self.test_step_outputs.clear()
        return

    def configure_optimizers(self):
        """
        Obtain optimizers and hyperparameter schedulers for model

        """
        optimizer = get_object(self.args.optimizer_name, "optimizer")(
            [p for p in self.parameters() if p.requires_grad], self.args
        )
        schedule = get_object(self.args.scheduler_name, "scheduler")(
            optimizer, self.args
        )

        scheduler = {
            "scheduler": schedule,
            "monitor": self.args.monitor,
            "interval": self.args.scheduler_interval,
            "frequency": 1,
        }
        return [optimizer], [scheduler]

    def compute_loss(self, model_output, batch):
        """
        Compute model loss:
            Iterates through loss functions defined in args and computes losses and predictions
            Adds losses and stores predictions for batch in dictionary

        Returns:
            total_loss (torch.Tensor): aggregate loss value that is propagated backwards for gradient computation
            logging_dict: dict of losses (and other metrics)
            predictions: dict of predictions (preds, probs, etc.)
        """
        total_loss = 0
        logging_dict, predictions = OrderedDict(), OrderedDict()
        for loss_fn in self.loss_fns[self.phase]:
            loss, l_dict, p_dict = loss_fn(model_output, batch, self, self.args)
            total_loss += loss
            logging_dict.update(l_dict)
            predictions.update(p_dict)
        return total_loss, logging_dict, predictions

    def call_metric(self, predictions, func):
        logging_dict = OrderedDict()
        for metric_fn in self.metrics[self.phase]:
            if func == "update":
                metric_fn.update(predictions, self.args)
            elif func == "compute":
                if metric_fn._update_called:
                    l_dict = metric_fn.compute()
                    logging_dict.update(l_dict)
        return logging_dict

    def store_in_predictions(self, preds, storage_dict):
        for m in get_object(self.args.dataset_name, "dataset").DATASET_ITEM_KEYS:
            if m in storage_dict:
                preds[m] = storage_dict[m]

        for m in self.metric_keys:
            if m in storage_dict:
                if torch.is_tensor(storage_dict[m]) and storage_dict[m].requires_grad:
                    preds[m] = storage_dict[m].detach()
                else:
                    preds[m] = storage_dict[m]
        return preds

    def log_outputs(self, outputs, key):
        """
        Compute performance metrics after epoch ends:
            Iterates through metric functions defined in args and computes metrics
            Logs the metric values into logger (Comet, Tensorboard, etc.)
        """
        logging_dict = {}
        for k, v in outputs.items():
            if isinstance(v, torch.Tensor) and any([i in k for i in self.LOG_KEYS]):
                logging_dict["{}_{}".format(key, k)] = v.mean()
        # log clocktime of methods for epoch
        if (self.args.profiler is not None) and (self.args.log_profiler):
            logging_dict.update(self.get_time_profile(key))
        self.log_dict(
            logging_dict,
            prog_bar=True,
            logger=True,
            batch_size=self.args.batch_size,
            sync_dist=True,
        )

    def get_time_profile(self, key):
        """Obtain trainer method times

        Args:
            key (str): one of ['train', 'val', 'test]

        Returns:
            dict: mean of clocktime of each method for past epoch
        """
        if key == "train":
            num_steps = self.trainer.num_training_batches
        if key == "val":
            num_steps = self.trainer.num_val_batches[0]
        if key == "test":
            num_steps = self.trainer.num_test_batches[0]

        time_profile = {}
        for k, v in self.trainer.profiler.recorded_durations.items():
            time_profile[k] = np.mean(v[-num_steps:])
        return time_profile

    def save_predictions(self, outputs):
        """
        Saves model predictions as pickle files
            Makes a directory under /inference_dir/experiment_name/
            Stores predictions for each sample individually under /inference_dir/experiment_name/sample_[sample_id].predictions

        * Requires outputs to contain the keys ['sample_id']
        """
        experiment_name = (
            os.path.splitext(os.path.basename(self.args.checkpoint_path))[0]
            if (self.args.from_checkpoint and not self.args.train)
            else self.args.experiment_name
        )
        for idx, sampleid in enumerate(outputs["sample_id"]):
            sampledict = {
                k: v[idx]
                for k, v in outputs.items()
                if (len(v) == len(outputs["sample_id"]))
            }
            for k, v in sampledict.items():
                if isinstance(v, torch.Tensor) and v.is_cuda:
                    sampledict[k] = v.cpu()
            predictions_filename = os.path.join(
                self.args.inference_dir,
                experiment_name,
                "sample_{}.predictions".format(sampleid),
            )
            dump_pickle(sampledict, predictions_filename)

    def save_hiddens(self, outputs):
        """
        Saves the model's hidden layer outputs as pickle files
            Makes a directory under /inference_dir/experiment_name/
            Stores predictions for each sample individually under /inference_dir/experiment_name/sample_[sample_id].hiddens

        * Requires outputs to contain the keys ['sample_id', 'hidden]
        """
        experiment_name = (
            os.path.splitext(os.path.basename(self.args.checkpoint_path))[0]
            if (self.args.from_checkpoint and not self.args.train)
            else self.args.experiment_name
        )
        idx = outputs["sample_id"]
        # hiddens = nn.functional.normalize(outputs['hidden'], dim = 1)
        hiddens = [
            {
                k: v[i].cpu() if v.is_cuda else v[i]
                for k, v in outputs.items()
                if ("hidden" in k) and (len(v) == len(idx))
            }
            for i in range(len(idx))
        ]
        for i, h in zip(idx, hiddens):
            predictions_filename = os.path.join(
                self.args.inference_dir, experiment_name, "sample_{}.hiddens".format(i)
            )
            dump_pickle(h, predictions_filename)

    def log_image(self, model_output, batch):
        # log one sample from each epoch
        sid = batch["sample_id"][0]
        for k, v in model_output.items():
            if "reconstruction" in k:
                img = model_output[k][0].detach().cpu()
                if img.shape[0] != 3:
                    img = img.numpy()
                    for cid, chan in enumerate(img):
                        self.logger.log_image(
                            chan,
                            "Sample{}_{}_Chan{}_Epoch{}_Step{}".format(
                                sid, k, cid, self.current_epoch, self.global_step
                            ),
                        )
                else:
                    img = img.permute(1, 2, 0).numpy()
                    self.logger.log_image(
                        img,
                        "Sample{}_{}_Epoch{}_Step{}".format(
                            sid, k, self.current_epoch, self.global_step
                        ),
                    )

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--model_name",
            type=str,
            action=set_nox_type("model"),
            default="classifier",
            help="Name of parent model",
        )
        parser.add_argument(
            "--compute_metrics_every_step",
            action="store_true",
            default=False,
            help="Whether to compute model metrics every step. Default is to compute at end of full epoch",
        )
        parser.add_argument(
            "--log_predictions_to_table",
            action="store_true",
            default=False,
            help="Whether to log predictions to a wandb table",
        )
        parser.add_argument(
            "--log_predictions_to_table_every",
            type=int,
            default=100,
            help="How often to log table predictions",
        )
        parser.add_argument(
            "--scheduler_interval",
            type=str,
            default="epoch",
            choices=["epoch", "step"],
            help="how often to apply scheduling on model parameters.",
        )


def gather_step_outputs(outputs):
    """
    Collates the dictionary outputs from each step into a single dictionary

    Returns:
        output_dict (dict): dictionary mapping step output keys to lists or tensors
    """

    output_dict = OrderedDict()
    if isinstance(outputs[-1], list):  # adversarial setting with two optimizers
        outputs = outputs[0]

    for k in outputs[-1].keys():
        if k == "preds_dict":
            output_dict[k] = gather_step_outputs(
                [output["preds_dict"] for output in outputs]
            )
        elif isinstance(outputs[-1][k], torch.Tensor):
            # check shapes are the same
            if not all(output[k].shape == outputs[-1][k].shape for output in outputs):
                output_dict[k] = [o for output in outputs for o in output[k]]
            elif len(outputs[-1][k].shape) == 0:
                output_dict[k] = torch.stack([output[k] for output in outputs])
            else:
                output_dict[k] = torch.cat([output[k] for output in outputs], dim=0)
        elif isinstance(outputs[-1][k], list):
            output_dict[k] = [olist for output in outputs for olist in output[k]]
        else:
            output_dict[k] = [output[k] for output in outputs]
    return output_dict


def dump_pickle(file_obj, file_name):
    """
    Saves object as a binary pickle file
        Creates directory of file
        Saves file

    Args:
        file_obj: object
        file_name: path to file
    """
    if not os.path.exists(os.path.dirname(file_name)):
        try:
            os.makedirs(os.path.dirname(file_name))
        except OSError as exc:  # Guard against race condition
            if exc.errno != exc.errno.EEXIST:
                raise
    pickle.dump(file_obj, open(file_name, "wb"))



# functions from torch_geometric.utils not yet importable
from rdkit import Chem, RDLogger
from typing import List
import torch
from torch import Tensor
import torch.nn.functional as F
from torch_geometric.data import Data
from torch_geometric.utils import degree
from itertools import combinations

x_map = {
    "atomic_num": list(range(0, 119)),
    "chirality": [
        "CHI_UNSPECIFIED",
        "CHI_TETRAHEDRAL_CW",
        "CHI_TETRAHEDRAL_CCW",
        "CHI_OTHER",
    ],
    "degree": list(range(0, 11)),
    "formal_charge": list(range(-5, 7)),
    "num_hs": list(range(0, 9)),
    "num_radical_electrons": list(range(0, 5)),
    "hybridization": [
        "UNSPECIFIED",
        "S",
        "SP",
        "SP2",
        "SP3",
        "SP3D",
        "SP3D2",
        "OTHER",
    ],
    "is_aromatic": [False, True],
    "is_in_ring": [False, True],
}

e_map = {
    "bond_type": [
        "misc",
        "SINGLE",
        "DOUBLE",
        "TRIPLE",
        "AROMATIC",
    ],
    "stereo": [
        "STEREONONE",
        "STEREOZ",
        "STEREOE",
        "STEREOCIS",
        "STEREOTRANS",
        "STEREOANY",
    ],
    "is_conjugated": [False, True],
}


def unbatch(src: Tensor, batch: Tensor, dim: int = 0) -> List[Tensor]:
    r"""Splits :obj:`src` according to a :obj:`batch` vector along dimension
    :obj:`dim`.

    Args:
        src (Tensor): The source tensor.
        batch (LongTensor): The batch vector
            :math:`\mathbf{b} \in {\{ 0, \ldots, B-1\}}^N`, which assigns each
            entry in :obj:`src` to a specific example. Must be ordered.
        dim (int, optional): The dimension along which to split the :obj:`src`
            tensor. (default: :obj:`0`)

    :rtype: :class:`List[Tensor]`
    """
    sizes = degree(batch, dtype=torch.long).tolist()
    return src.split(sizes, dim)


def from_smiles(
    smiles: str,
    with_hydrogen: bool = False,
    kekulize: bool = False,
    return_atom_number=False,
    use_one_hot_encoding=False,
):
    r"""Converts a SMILES string to a :class:`torch_geometric.data.Data`
    instance.
    Args:
        smiles (string, optional): The SMILES string.
        with_hydrogen (bool, optional): If set to :obj:`True`, will store
            hydrogens in the molecule graph. (default: :obj:`False`)
        kekulize (bool, optional): If set to :obj:`True`, converts aromatic
            bonds to single/double bonds. (default: :obj:`False`)
    """

    RDLogger.DisableLog("rdApp.*")

    mol = Chem.MolFromSmiles(smiles)

    if mol is None:
        mol = Chem.MolFromSmiles("")
    if with_hydrogen:
        mol = Chem.AddHs(mol)
    if kekulize:
        mol = Chem.Kekulize(mol)

    # mapping order = from2to
    xs = []
    x_ids = []
    x_numbers = []
    for atom in mol.GetAtoms():
        x = []
        x.append(x_map["atomic_num"].index(atom.GetAtomicNum()))
        x.append(x_map["chirality"].index(str(atom.GetChiralTag())))
        x.append(x_map["degree"].index(atom.GetTotalDegree()))
        x.append(x_map["formal_charge"].index(atom.GetFormalCharge()))
        x.append(x_map["num_hs"].index(atom.GetTotalNumHs()))
        x.append(
            x_map["num_radical_electrons"].index(max(atom.GetNumRadicalElectrons(), 4))
        )
        x.append(x_map["hybridization"].index(str(atom.GetHybridization())))
        x.append(x_map["is_aromatic"].index(atom.GetIsAromatic()))
        x.append(x_map["is_in_ring"].index(atom.IsInRing()))
        xs.append(x)
        x_ids.append(atom.GetIdx())
        if return_atom_number:
            x_numbers.append(
                int(atom.GetProp("molAtomMapNumber"))
                if atom.HasProp("molAtomMapNumber")
                else -1
            )

    x = torch.tensor(xs, dtype=torch.long).view(-1, 9)
    # x_ids = torch.tensor(x_ids, dtype=torch.long).view(-1)

    # Map i and j to the new order
    edge_indices, edge_attrs = [], []
    for bond in mol.GetBonds():
        i = bond.GetBeginAtomIdx()
        j = bond.GetEndAtomIdx()

        e = []
        e.append(e_map["bond_type"].index(str(bond.GetBondType())))
        e.append(e_map["stereo"].index(str(bond.GetStereo())))
        e.append(e_map["is_conjugated"].index(bond.GetIsConjugated()))

        edge_indices += [[i, j], [j, i]]
        edge_attrs += [e, e]

    edge_index = torch.tensor(edge_indices)
    edge_index = edge_index.t().to(torch.long).view(2, -1)
    edge_attr = torch.tensor(edge_attrs, dtype=torch.long).view(-1, 3)

    if edge_index.numel() > 0:  # Sort indices.
        perm = (edge_index[0] * x.size(0) + edge_index[1]).argsort()
        edge_index, edge_attr = edge_index[:, perm], edge_attr[perm]

    if use_one_hot_encoding:
        one_hot_x = torch.hstack(
            [F.one_hot(x[:, i], len(v)) for i, (k, v) in enumerate(x_map.items())]
        )
        if x.shape[-1] > len(x_map):
            one_hot_x = torch.hstack([one_hot_x, x[:, len(x_map) :]])
        x = one_hot_x

        one_hot_e = torch.hstack(
            [
                F.one_hot(edge_attr[:, i], len(v))
                for i, (k, v) in enumerate(e_map.items())
            ]
        )
        if edge_attr.shape[-1] > len(e_map):
            one_hot_e = torch.hstack([one_hot_e, x[:, len(e_map) :]])
        edge_attr = one_hot_e

    data = Data(
        x=x, edge_index=edge_index, edge_attr=edge_attr, smiles=smiles, x_ids=x_ids
    )
    if return_atom_number:
        data.atom_map_number = x_numbers  # torch.tensor(x_numbers).view(-1)
    return data


def from_mapped_smiles(
    smiles: str,
    with_hydrogen: bool = False,
    kekulize: bool = False,
    encode_no_edge=False,
    sanitize=True,
    use_one_hot_encoding=False,
):
    r"""Converts a SMILES string to a :class:`torch_geometric.data.Data`
    instance.
    Args:
        smiles (string, optional): The SMILES string.
        with_hydrogen (bool, optional): If set to :obj:`True`, will store
            hydrogens in the molecule graph. (default: :obj:`False`)
        kekulize (bool, optional): If set to :obj:`True`, converts aromatic
            bonds to single/double bonds. (default: :obj:`False`)
        encode_no_edge (bool, optional): adds edges between all molecules
        sanitize (bool, optional): sanitize molecules when using rdkit. typically should be set to `True`
    """
    RDLogger.DisableLog("rdApp.*")

    mol = Chem.MolFromSmiles(smiles, sanitize=sanitize)

    if mol is None:
        return None, None  # mol = Chem.MolFromSmiles("")
    if with_hydrogen:
        mol = Chem.AddHs(mol)
    if kekulize:
        mol = Chem.Kekulize(mol)

    xs = []
    x_ids = []
    x_numbers = []
    atom_index_map = {}
    for atom in mol.GetAtoms():
        x = []
        x.append(x_map["atomic_num"].index(atom.GetAtomicNum()))
        x.append(x_map["chirality"].index(str(atom.GetChiralTag())))
        x.append(x_map["degree"].index(atom.GetTotalDegree()))
        x.append(x_map["formal_charge"].index(atom.GetFormalCharge()))
        x.append(x_map["num_hs"].index(atom.GetTotalNumHs()))
        x.append(
            x_map["num_radical_electrons"].index(max(atom.GetNumRadicalElectrons(), 4))
        )
        x.append(x_map["hybridization"].index(str(atom.GetHybridization())))
        x.append(x_map["is_aromatic"].index(atom.GetIsAromatic()))
        x.append(x_map["is_in_ring"].index(atom.IsInRing()))
        xs.append(x)
        x_ids.append(atom.GetIdx())

        assert atom.HasProp("molAtomMapNumber"), "SMILES must contain atom map numbers"

        atom_map_number = int(atom.GetProp("molAtomMapNumber"))
        x_numbers.append(atom_map_number)
        atom_index_map[atom.GetIdx()] = atom_map_number  # Update mapping dictionary

    new_xs = [None for _ in range(len(xs))]
    order = sorted(
        atom_index_map.items(), key=lambda x: x[1]
    )  # sort by atom number because k,v = atom_idx, atom_map_number
    old_index2new_index = {
        old_idx: new_idx for new_idx, (old_idx, _) in enumerate(order)
    }
    atom_map_number2new_index = {
        atom_map_number: new_idx for new_idx, (_, atom_map_number) in enumerate(order)
    }

    for i, (atom_idx, atom_map_number) in enumerate(order):
        new_xs[i] = xs[atom_idx]

    x = torch.tensor(new_xs, dtype=torch.long).view(-1, 9)

    edge_indices, edge_attrs = [], []
    edge_indices_complete, edge_attrs_complete = [], []
    for bond in mol.GetBonds():
        i = bond.GetBeginAtomIdx()
        j = bond.GetEndAtomIdx()

        # Remap i and j according to the atom_map_number
        i = old_index2new_index[i]
        j = old_index2new_index[j]

        e = []
        e_complete = [0, 1]  # [different components, same component, bond features]

        e.append(e_map["bond_type"].index(str(bond.GetBondType())))
        e.append(e_map["stereo"].index(str(bond.GetStereo())))
        e.append(e_map["is_conjugated"].index(bond.GetIsConjugated()))

        e_complete.append(
            e_map["bond_type"].index(str(bond.GetBondType())) + 1
        )  # ! add 1 in case we use encode_no_edge
        e_complete.append(
            e_map["stereo"].index(str(bond.GetStereo())) + 1
        )  # ! add 1 in case we use encode_no_edge
        e_complete.append(
            e_map["is_conjugated"].index(bond.GetIsConjugated()) + 1
        )  # ! add 1 in case we use encode_no_edge

        edge_indices += [[i, j], [j, i]]
        edge_attrs += [e, e]
        edge_indices_complete += [[i, j], [j, i]]
        edge_attrs_complete += [e_complete, e_complete]

    if encode_no_edge:
        # atom id to molecule id (if multiple in smiles)
        atom2molecule = {}
        for mol_id, s in enumerate(smiles.split(".")):
            single_mol = Chem.MolFromSmiles(s, sanitize=sanitize)
            for atom in single_mol.GetAtoms():
                # idx = old_index2new_index[ atom.GetIdx() ] # ! problem: atom.GetIdx() gets reset for each smiles in list; use atom number
                idx = atom_map_number2new_index[int(atom.GetProp("molAtomMapNumber"))]
                atom2molecule[idx] = mol_id  # new index to molecule id

        edge_attr_dim = len(e_map)
        set_edge_indices = set([tuple(idxs) for idxs in edge_indices])
        total_num_atoms = mol.GetNumAtoms()
        for i, j in combinations(
            range(total_num_atoms), 2
        ):  # get all combs of new indices
            if (i, j) not in set_edge_indices:
                if atom2molecule[i] == atom2molecule[j]:  # same molecule (component)
                    e = [0, 1] + [0 for _ in range(edge_attr_dim)]
                else:
                    e = [1, 0] + [0 for _ in range(edge_attr_dim)]
                edge_indices_complete += [[i, j], [j, i]]
                edge_attrs_complete += [e, e]
                set_edge_indices.update(
                    [(i, j), (j, i)]
                )  # prob not necessary but just to be safe

    edge_index = torch.tensor(edge_indices)
    edge_index = edge_index.t().to(torch.long).view(2, -1)
    edge_attr = torch.tensor(edge_attrs, dtype=torch.long).view(-1, 3)

    edge_index_complete = torch.tensor(edge_indices_complete)
    edge_index_complete = edge_index_complete.t().to(torch.long).view(2, -1)
    edge_attr_complete = torch.tensor(edge_attrs_complete, dtype=torch.long).view(-1, 5)

    if edge_index.numel() > 0:  # Sort indices.
        perm = (edge_index[0] * x.size(0) + edge_index[1]).argsort()
        edge_index, edge_attr = edge_index[:, perm], edge_attr[perm]

        perm = (edge_index_complete[0] * x.size(0) + edge_index_complete[1]).argsort()
        edge_index_complete, edge_attr_complete = (
            edge_index_complete[:, perm],
            edge_attr_complete[perm],
        )

    if use_one_hot_encoding:
        one_hot_x = torch.hstack(
            [F.one_hot(x[:, i], len(v)) for i, (k, v) in enumerate(x_map.items())]
        )
        if x.shape[-1] > len(x_map):
            one_hot_x = torch.hstack([one_hot_x, x[:, len(x_map) :]])
        x = one_hot_x

        one_hot_e = torch.hstack(
            [
                F.one_hot(edge_attr[:, i], len(v))
                for i, (k, v) in enumerate(e_map.items())
            ]
        )
        if edge_attr.shape[-1] > len(e_map):
            one_hot_e = torch.hstack([one_hot_e, x[:, len(e_map) :]])
        edge_attr = one_hot_e

    if encode_no_edge:
        if use_one_hot_encoding:
            one_hot_e = torch.hstack(
                [
                    F.one_hot(edge_attr_complete[:, (i + 2)], len(v) + 1)
                    for i, (k, v) in enumerate(e_map.items())
                ]
            )  # start at i = 2 since first 2 are for encoding same or diff mol
            one_hot_e = torch.hstack([edge_attr_complete[:, :2], one_hot_e])
            edge_attr_complete = one_hot_e

        data = Data(
            x=x,
            edge_index=edge_index,
            edge_attr=edge_attr,
            smiles=smiles,
            x_ids=x_ids,
            edge_index_complete=edge_index_complete,
            edge_attr_complete=edge_attr_complete,
        )
    else:
        data = Data(
            x=x, edge_index=edge_index, edge_attr=edge_attr, smiles=smiles, x_ids=x_ids
        )
    data.atom_map_number = x_numbers  # torch.tensor(x_numbers).view(-1)
    return data, atom_map_number2new_index



import torch
from torch import nn, einsum, broadcast_tensors
import torch.nn.functional as F
from torch.nn import SiLU

from clipzyme.utils.registry import register_object, get_object
from clipzyme.models.abstract import AbstractModel
from clipzyme.utils.classes import set_nox_type

from einops import rearrange, repeat
from einops.layers.torch import Rearrange

import copy
import os, sys
import pickle
import warnings
import hashlib
from collections import defaultdict
from rich import print as rprint
from typing import Optional, List, Union, Tuple, Any, Dict

import torch_geometric
from torch_geometric.nn import MessagePassing
from torch_geometric.typing import Adj, Size, OptTensor, Tensor
from torch_geometric.data import HeteroData, Data, Batch, Dataset
from torch_scatter import scatter
from torch_geometric.utils import to_dense_batch
from torch_geometric.utils.loop import add_remaining_self_loops

import math


def exists(val):
    return val is not None


class CoorsNorm(nn.Module):
    def __init__(self, eps=1e-8, scale_init=1.0):
        super().__init__()
        self.eps = eps
        scale = torch.zeros(1).fill_(scale_init)
        self.scale = nn.Parameter(scale)

    def forward(self, coors):
        norm = coors.norm(dim=-1, keepdim=True)
        normed_coors = coors / norm.clamp(min=self.eps)
        return normed_coors * self.scale


class SinusoidalEmbeddings(nn.Module):
    """A simple sinusoidal embedding layer. From Jeremy."""

    def __init__(self, dim, theta: float = 10000.0) -> None:
        super().__init__()
        self.dim = dim
        self.theta = theta

    def forward(self, time: torch.Tensor) -> torch.Tensor:
        half_dim = self.dim // 2
        embeddings = math.log(self.theta) / (half_dim - 1)
        embeddings = torch.exp(torch.arange(half_dim, device=time.device) * -embeddings)
        embeddings = time[:, None] * embeddings[None, :]
        embeddings = torch.cat((embeddings.sin(), embeddings.cos()), dim=-1)

        return embeddings


class EGNN_Sparse(MessagePassing):
    """ """

    propagate_type = {
        "x": Tensor,
        "edge_attr": Tensor,
        "coors": Tensor,
        "rel_coors": Tensor,
        "size": Size,
        "batch": Dict,
    }

    def __init__(self, args, **kwargs):
        self.args = args
        aggr = self.args.neighbour_aggr
        self.update_feats = self.args.update_feats
        self.update_coors = self.args.update_coors
        assert aggr in {
            "add",
            "sum",
            "max",
            "mean",
        }, "pool method must be a valid option"
        assert (
            self.update_feats or self.update_coors
        ), "you must update either features, coordinates, or both"
        kwargs.setdefault("aggr", aggr)
        super(EGNN_Sparse, self).__init__(**kwargs)
        # model params
        self.feats_dim = self.args.protein_dim
        self.feat_proj_dim = self.args.feat_proj_dim
        self.edge_attr_dim = self.args.edge_attr_dim
        self.m_dim = self.args.message_dim
        self.soft_edge = self.args.soft_edge
        self.norm_feats = self.args.norm_feats
        self.norm_coors = self.args.norm_coors
        self.norm_coors_scale_init = self.args.norm_coors_scale_init
        self.coor_weights_clamp_value = self.args.coor_weights_clamp_value

        if not self.args.use_sinusoidal:
            self.edge_input_dim = self.edge_attr_dim + 1 + (self.feats_dim * 2)
        else:
            self.edge_input_dim = self.feats_dim * 3
        self.dropout = (
            nn.Dropout(self.args.dropout) if self.args.dropout > 0 else nn.Identity()
        )

        dist_dim = (
            self.args.protein_dim
        )  # can replace if using different distance embedding
        if self.args.use_sinusoidal:
            self.dist_embedding = SinusoidalEmbeddings(dist_dim)
        else:
            self.dist_embedding = nn.Linear(1, dist_dim)  # type: ignore

        # EDGES
        # \phi_e
        self.edge_mlp = nn.Sequential(
            nn.Linear(self.edge_input_dim, self.edge_input_dim * 2),
            self.dropout,
            SiLU(),
            nn.Linear(self.edge_input_dim * 2, self.m_dim),
            SiLU(),
        )

        self.edge_weight = (
            nn.Sequential(nn.Linear(self.m_dim, 1), nn.Sigmoid())
            if self.soft_edge
            else None
        )

        self.node_norm = (
            torch_geometric.nn.norm.LayerNorm(self.feats_dim)
            if self.norm_feats
            else None
        )
        self.coors_norm = (
            CoorsNorm(scale_init=self.norm_coors_scale_init)
            if self.norm_coors
            else nn.Identity()
        )

        # \phi_h
        self.node_mlp = (
            nn.Sequential(
                nn.Linear(self.feats_dim + self.m_dim, self.feat_proj_dim * 2),
                self.dropout,
                SiLU(),
                nn.Linear(self.feat_proj_dim * 2, self.feats_dim),
            )
            if self.update_feats
            else None
        )

        # COORS
        # \phi_x
        self.coors_mlp = (
            nn.Sequential(
                nn.Linear(self.m_dim, self.m_dim * 4),
                self.dropout,
                SiLU(),
                nn.Linear(self.m_dim * 4, 1),
            )
            if self.update_coors
            else None
        )

        self.apply(self.init_)

    def init_(self, module):
        if type(module) in {nn.Linear}:
            # seems to be needed to keep the network from exploding to NaN with greater depths
            nn.init.xavier_normal_(module.weight)
            nn.init.zeros_(module.bias)

    def forward(
        self,
        feats: Tensor,
        coors: Tensor,
        edge_index: Adj,
        edge_attr: OptTensor = None,
        batch: Adj = None,
        size: Size = None,
    ) -> Tensor:
        """ """
        rel_coors = coors[edge_index[0]] - coors[edge_index[1]]
        # rel_dist = (rel_coors**2).sum(dim=-1, keepdim=True)
        rel_dist = torch.sqrt((rel_coors**2).sum(dim=-1, keepdim=True))
        rel_dist = self.dist_embedding(rel_dist)
        if rel_dist.shape[1] == 1:
            rel_dist = rel_dist.squeeze(1)

        if exists(edge_attr):
            edge_attr_feats = torch.cat([edge_attr, rel_dist], dim=-1)
        else:
            edge_attr_feats = rel_dist

        hidden_out, coors_out = self.propagate(
            edge_index,
            x=feats,
            edge_attr=edge_attr_feats,
            coors=coors,
            rel_coors=rel_coors,
            size=None,
            batch=batch,
        )
        return coors_out, hidden_out

    def message(self, x_i, x_j, edge_attr) -> Tensor:
        m_ij = self.edge_mlp(torch.cat([x_i, x_j, edge_attr], dim=-1))
        return m_ij

    def propagate(self, edge_index: Adj, size: Size = None, **kwargs):
        """
        The initial call to start propagating messages.

        Args:
            `edge_index` holds the indices of a general (sparse)
                assignment matrix of shape :obj:`[N, M]`.
            size (tuple, optional) if none, the size will be inferred
                and assumed to be quadratic.
            **kwargs: Any additional data which is needed to construct and
                aggregate messages, and to update node embeddings.
        """
        # try:
        #     size = self.__check_input__(edge_index, size)
        #     coll_dict = self.__collect__(self.__user_args__, edge_index, size, kwargs)
        # except:
        size = self._check_input(edge_index, size)
        coll_dict = self._collect(self._user_args, edge_index, size, kwargs)
        msg_kwargs = self.inspector.distribute("message", coll_dict)
        aggr_kwargs = self.inspector.distribute("aggregate", coll_dict)
        update_kwargs = self.inspector.distribute("update", coll_dict)

        # get messages
        m_ij = self.message(**msg_kwargs)  # graph_size x m_dim

        # update coors if specified
        if self.update_coors:
            coor_wij = self.coors_mlp(m_ij)  # graph_size x 1
            # clamp if arg is set
            if self.coor_weights_clamp_value:
                clamp_value = self.coor_weights_clamp_value
                coor_weights.clamp_(min=-clamp_value, max=clamp_value)

            # normalize if needed
            kwargs["rel_coors"] = self.coors_norm(kwargs["rel_coors"])  # graph_size x 3

            mhat_i = self.aggregate(
                coor_wij * kwargs["rel_coors"], **aggr_kwargs
            )  # graph_size / num_neighbours x 3
            coors_out = kwargs["coors"] + mhat_i  # graph_size / num_neighbours x 3
        else:
            coors_out = kwargs["coors"]

        # update feats if specified
        if self.update_feats:
            # weight the edges if arg is passed
            if self.soft_edge:
                m_ij = m_ij * self.edge_weight(m_ij)
            m_i = self.aggregate(m_ij, **aggr_kwargs)

            hidden_feats = (
                self.node_norm(kwargs["x"], kwargs["batch"])
                if self.node_norm
                else kwargs["x"]
            )
            hidden_out = self.node_mlp(
                torch.cat([hidden_feats, m_i], dim=-1)
            )  # graph_size / num_neighbours x feat_dim (esm)

            hidden_out = kwargs["x"] + hidden_out
        else:
            hidden_out = kwargs["x"]

        return self.update((hidden_out, coors_out), **update_kwargs)


@register_object("egnn_sparse_network", "model")
class EGNN_Sparse_Network(AbstractModel):
    """ """

    def __init__(self, args):
        super(EGNN_Sparse_Network, self).__init__()
        self.args = args
        self.mpnn_layers = nn.ModuleList()

        for i in range(self.args.egcl_layers):
            layer = EGNN_Sparse(args)
            self.mpnn_layers.append(layer)

    def forward(self, batch):
        """ """
        if (
            hasattr(batch, "graph")
            or "graph" in batch
            or (
                isinstance(batch, (Data, HeteroData))
                and "graph" in batch.to_dict().keys()
            )
        ):
            coors = batch["graph"]["receptor"].pos
            feats = batch["graph"]["receptor"].x
            edge_index = batch["graph"]["receptor", "contact", "receptor"].edge_index
            batch_idx = batch["graph"]["receptor"].batch
        elif (
            hasattr(batch, "receptor")
            or "receptor" in batch
            or (
                isinstance(batch, (Data, HeteroData))
                and "receptor" in batch.to_dict().keys()
            )
        ):
            coors = batch["receptor"].pos
            feats = batch["receptor"].x
            edge_index = batch["receptor", "contact", "receptor"].edge_index
            batch_idx = batch["receptor"].batch
        else:
            raise ValueError("batch must have a graph or receptor attribute")

        edge_attr = None
        bsize = batch_idx.max() + 1

        for i, egcl in enumerate(self.mpnn_layers):
            coors, feats = egcl(feats, coors, edge_index, size=bsize, batch=batch_idx)

        return feats, coors

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--pool_type",
            type=str,
            choices=["none", "sum", "mul", "mean", "min", "max"],
            default="none",
            help="Type of pooling to do to obtain graph features",
        )
        parser.add_argument(
            "--neighbour_aggr",
            type=str,
            choices=["add", "sum", "mean", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )
        parser.add_argument(
            "--edge_attr_dim",
            type=int,
            default=0,
            help="",
        )
        parser.add_argument(
            "--message_dim",
            type=int,
            default=16,
            help="",
        )
        parser.add_argument(
            "--soft_edge",
            type=int,
            default=0,
            help="",
        )
        parser.add_argument(
            "--norm_feats",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--norm_coors",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--update_feats",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--update_coors",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--norm_coors_scale_init",
            type=float,
            default=1e-2,
            help="",
        )
        parser.add_argument(
            "--coor_weights_clamp_value",
            type=float,
            default=None,
            help="",
        )
        parser.add_argument(
            "--egcl_layers",
            type=int,
            default=4,
            help="",
        )
        parser.add_argument(
            "--protein_dim",
            type=int,
            default=64,
            help="",
        )
        parser.add_argument(
            "--feat_proj_dim",
            type=int,
            default=1280,
            help="dim that features are projected to upon being fed to phi_e",
        )
        parser.add_argument(
            "--use_sinusoidal",
            action="store_true",
            default=False,
            help="whether or not to use sinusoidal embeddings for distance",
        )


@register_object("egnn_classifier", "model")
class EGNN_Classifier(AbstractModel):
    def __init__(self, args):
        super(EGNN_Classifier, self).__init__()

        self.args = args

        self.egnn = EGNN_Sparse_Network(args)

        self.esm_model = get_object(args.esm_model_name, "model")(args)
        classifier_args = copy.deepcopy(args)
        self.mlp = get_object(args.classifier_name, "model")(classifier_args)

    def forward(self, batch=None):
        output = {}
        # run ESM and update node embeddings
        batch = self.compute_embeddings(batch, self.args)
        del batch["mask_hiddens"]  # not used and is big
        feats, coors = self.egnn(batch)
        output["node_features"] = feats
        output["node_coords"] = coors
        if self.args.pool_type != "none":
            output["graph_features"] = scatter(
                feats,
                batch["graph"]["receptor"].batch,
                dim=0,
                reduce=self.args.pool_type,
            )
        else:
            output["graph_features"] = output["node_features"]

        output.update(self.mlp({"x": output["graph_features"]}))

        return output

    def compute_embeddings(self, batch, args):
        """
        Pre-compute ESM2 embeddings.
        """
        batch_hash = hashlib.md5(f"{batch}".encode()).hexdigest()
        esm_cache = os.path.join(
            args.protein_cache_path, f"{batch_hash}_cached_prot_esm.pkl"
        )
        if args.debug:
            esm_cache = esm_cache.replace(".pkl", "_debug.pkl")

        # check if we already computed embeddings
        if not args.no_graph_cache and os.path.exists(esm_cache):
            with open(esm_cache, "rb") as f:
                cropped_representations = pickle.load(f)
            self._save_esm_rep(batch, cropped_representations)
            print("Loaded cached ESM embeddings")
            return batch

        # run ESM model, use args.freeze_esm for no training
        esm_output = self.esm_model(batch)

        tokens = esm_output["tokens"]
        reps = esm_output["token_hiddens"]

        cropped_representations = []
        seq_len = [len(p) for p in batch["sequence"]]

        # TODO: use padding mask
        for i, rep in enumerate(reps):
            length = seq_len[i]
            rep_cropped = rep[
                1 : length + 1
            ]  # removes cls at idx 0 and removes eos at idx length + 2
            token_cropped = tokens[i][1 : length + 1, None]
            cropped_representations.append((rep_cropped, token_cropped))

        # dump to cache
        if not args.no_graph_cache:
            with open(esm_cache, "wb+") as f:
                pickle.dump(cropped_representations, f)

        # overwrite graph.x for each element in batch
        self._save_esm_rep(batch, cropped_representations)

        batch["mask_hiddens"] = esm_output["mask_hiddens"]

        return batch

    def _save_esm_rep(self, batch, cropped_representations):
        """
        Assign new ESM representation to graph.x INPLACE
        """
        all_graphs = batch["graph"].to_data_list()
        # cropped_representations = (rep_crop, token_crop)
        for idx, graph in enumerate(all_graphs):
            rec_graph = graph["receptor"]
            rec_graph.x = cropped_representations[idx][0]  # esm embedding
            assert len(rec_graph.pos) == len(rec_graph.x)

        batch["graph"] = Batch.from_data_list(all_graphs, None, None)

        return batch

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--classifier_name",
            type=str,
            action=set_nox_type("model"),
            default="mlp_classifier",
            help="Name of encoder to use",
        )
        parser.add_argument(
            "--esm_model_name",
            type=str,
            action=set_nox_type("model"),
            default="fair_esm2",
            help="Name of esm encoder to use",
        )
        parser.add_argument(
            "--pool_type",
            type=str,
            choices=["none", "sum", "mul", "mean", "min", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )
        parser.add_argument(
            "--neighbour_aggr",
            type=str,
            choices=["add", "sum", "mean", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )
        parser.add_argument(
            "--edge_attr_dim",
            type=int,
            default=0,
            help="",
        )
        parser.add_argument(
            "--message_dim",
            type=int,
            default=16,
            help="",
        )
        parser.add_argument(
            "--soft_edge",
            type=int,
            default=0,
            help="",
        )
        parser.add_argument(
            "--norm_feats",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--norm_coors",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--update_feats",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--update_coors",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--norm_coors_scale_init",
            type=float,
            default=1e-2,
            help="",
        )
        parser.add_argument(
            "--coor_weights_clamp_value",
            type=float,
            default=None,
            help="",
        )
        parser.add_argument(
            "--egcl_layers",
            type=int,
            default=4,
            help="",
        )


@register_object("egnn_active_site_classifier", "model")
class EGNN_Active_Site_Classifier(EGNN_Classifier):
    def __init__(self, args):
        super(EGNN_Active_Site_Classifier, self).__init__(args)

    def forward(self, batch=None):
        output = {}
        # run ESM and update node embeddings
        batch = self.compute_embeddings(batch, self.args)
        feats, coors = self.egnn(batch)
        output["node_features"] = feats
        output["node_coords"] = coors
        # reshape to batch x nodes x features
        output["graph_features"], mask = to_dense_batch(
            output["node_features"], batch=batch["graph"]["receptor"].batch
        )
        output.update(
            self.mlp({"x": output["graph_features"]})
        )  # batch x nodes x features -> batch x nodes x 1

        output["logit"] = output["logit"].squeeze(-1)
        labels = batch["residue_mask"]
        residue_mask = batch["mask_hiddens"][:, 1:-1]
        # cross entropy will ignore the padded residues (ignore_index=-100)
        labels[~residue_mask.squeeze(-1).bool()] = -100
        batch["y"] = labels

        return output

    @staticmethod
    def set_args(args) -> None:
        super(EGNN_Active_Site_Classifier, EGNN_Active_Site_Classifier).set_args(args)
        args.num_classes = 2


@register_object("egnn_active_site_conv_classifier", "model")
class EGNN_Active_Site_Conv_Classifier(EGNN_Active_Site_Classifier):
    def __init__(self, args):
        super(EGNN_Active_Site_Conv_Classifier, self).__init__(args)

    def forward(self, batch=None):
        output = {}
        # run ESM and update node embeddings
        batch = self.compute_embeddings(batch, self.args)
        feats, coors = self.egnn(batch)
        output["node_features"] = feats
        output["node_coords"] = coors
        output.update(self.mlp({"x": feats}))  # nodes x 1

        # nodes x 1
        aggr_logits = self.aggregate_neighborhood(
            output["logit"].squeeze(-1),
            batch["graph"]["receptor", "contact", "receptor"].edge_index,
        )

        # reshape to batch x nodes x 1
        logit, mask = to_dense_batch(
            aggr_logits, batch=batch["graph"]["receptor"].batch
        )

        output["logit"] = logit

        labels = batch[
            "residue_mask"
        ]  # .reshape(1, -1).squeeze(0) # flatten -> num nodes x 1
        cropped_labels = []
        for i, seq_label in enumerate(labels):
            seq_len = len(batch["sequence"][i])
            label_crop = seq_label[:seq_len]
            cropped_labels.append(label_crop)

        cropped_labels = torch.cat(cropped_labels)
        aggr_labels = self.aggregate_neighborhood(
            cropped_labels, batch["graph"]["receptor", "contact", "receptor"].edge_index
        )  # num nodes x 1
        labels, labels_mask = to_dense_batch(
            aggr_labels.unsqueeze(-1), batch=batch["graph"]["receptor"].batch
        )  # batch x nodes x 1

        residue_mask = batch["mask_hiddens"][:, 1:-1]
        # cross entropy will ignore the padded residues (ignore_index=-100)
        labels[~residue_mask.squeeze(-1).bool()] = -100
        batch["y"] = labels.squeeze(-1)
        return output

    def aggregate_neighborhood(self, logits, edge_index, reduce="max"):
        edge_idx_self_loops, edge_attr_self_loops = add_remaining_self_loops(edge_index)
        aggr_logits = scatter(
            logits[edge_idx_self_loops[0]], edge_idx_self_loops[1], reduce=reduce
        )
        return aggr_logits

    @staticmethod
    def set_args(args) -> None:
        super(
            EGNN_Active_Site_Conv_Classifier, EGNN_Active_Site_Conv_Classifier
        ).set_args(args)
        args.num_classes = 1


@register_object("egnn_mol_classifier", "model")
class EGNN_Mol_Classifier(AbstractModel):
    def __init__(self, args):
        super(EGNN_Mol_Classifier, self).__init__()

        self.args = args
        self.egnn = EGNN_Sparse_Network(args)
        self.chemprop = get_object(args.chemprop_name, "model")(args)
        classifier_args = copy.deepcopy(args)
        classifier_args.mlp_input_dim = (
            classifier_args.protein_dim + args.chemprop_hidden_dim
        )
        if self.args.use_rdkit_features:
            classifier_args.mlp_input_dim = (
                args.graph_classifier_hidden_dim + args.rdkit_features_dim
            )
        self.mlp = get_object(args.classifier_name, "model")(classifier_args)

    def forward(self, batch=None):
        output = {}
        # get protein features
        feats, coors = self.egnn(batch)
        if self.args.pool_type != "none":
            output["prot_features"] = scatter(
                feats,
                batch["receptor"].batch,
                dim=0,
                reduce=self.args.pool_type,
            )
        else:
            output["prot_features"] = feats

        # get molecule features
        if isinstance(batch["mol_data"], list):
            mol_output = self.chemprop(Batch.from_data_list(batch["mol_data"]))
        elif isinstance(batch["mol_data"], Batch):
            mol_output = self.chemprop(batch["mol_data"])
        else:
            raise ValueError(
                "mol_data must be a list of Data/HeteroData objs or Batch object"
            )
        output["mol_features"] = mol_output["graph_features"]
        if self.args.use_rdkit_features:
            features = batch["rdkit_features"].view(batch_size, -1)
            output["mol_features"] = torch.concat(
                [output["mol_features"], features], dim=-1
            )
        # concatenate and classify
        output.update(
            self.mlp(
                {
                    "x": torch.cat(
                        [output["prot_features"], output["mol_features"]], dim=-1
                    )
                }
            )
        )

        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--classifier_name",
            type=str,
            action=set_nox_type("model"),
            default="mlp_classifier",
            help="Name of encoder to use",
        )
        parser.add_argument(
            "--chemprop_name",
            type=str,
            action=set_nox_type("model"),
            default="chemprop",
            help="Name of mol encoder to use",
        )
        parser.add_argument(
            "--pool_type",
            type=str,
            choices=["none", "sum", "mul", "mean", "min", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )
        parser.add_argument(
            "--neighbour_aggr",
            type=str,
            choices=["add", "sum", "mean", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )
        parser.add_argument(
            "--edge_attr_dim",
            type=int,
            default=0,
            help="",
        )
        parser.add_argument(
            "--message_dim",
            type=int,
            default=16,
            help="",
        )
        parser.add_argument(
            "--soft_edge",
            type=int,
            default=0,
            help="",
        )
        parser.add_argument(
            "--norm_feats",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--norm_coors",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--update_feats",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--update_coors",
            action="store_true",
            default=False,
            help="",
        )
        parser.add_argument(
            "--norm_coors_scale_init",
            type=float,
            default=1e-2,
            help="",
        )
        parser.add_argument(
            "--coor_weights_clamp_value",
            type=float,
            default=None,
            help="",
        )
        parser.add_argument(
            "--egcl_layers",
            type=int,
            default=4,
            help="",
        )
        parser.add_argument(
            "--protein_dim",
            type=int,
            default=64,
            help="",
        )
        parser.add_argument(
            "--feat_proj_dim",
            type=int,
            default=1280,
            help="dim that features are projected to upon being fed to phi_e",
        )
        parser.add_argument(
            "--use_sinusoidal",
            action="store_true",
            default=False,
            help="whether or not to use sinusoidal embeddings for distance",
        )
        parser.add_argument(
            "--use_rdkit_features",
            action="store_true",
            default=False,
            help="whether using graph-level features from rdkit",
        )
        parser.add_argument(
            "--rdkit_features_dim",
            type=int,
            default=0,
            help="number of features",
        )



from argparse import Namespace, FileType
import pickle
import collections.abc as container_abcs
import re
from tabnanny import check
from typing import Literal, Optional
from clipzyme.utils.registry import get_object
import torch
from torch.utils import data
from clipzyme.utils.sampler import DistributedWeightedSampler

try:
    from pytorch_lightning.utilities.cloud_io import load as pl_load
except:
    from lightning_fabric.utilities.cloud_io import _load as pl_load
from torch_geometric.data import Data, HeteroData, Batch
from rich import print as rprint

string_classes = (str, bytes)
int_classes = int
np_str_obj_array_pattern = re.compile(r"[SaUO]")

default_collate_err_msg_format = (
    "default_collate: batch must contain tensors, numpy arrays, numbers, PyG Data or HeteroData, "
    "dicts, or lists; found {}"
)


def default_collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, Batch):
        return batch
    if isinstance(elem, (Data, HeteroData)):
        # optional args for more complex graphs (see pytorch geometric https://pytorch-geometric.readthedocs.io/en/latest/notes/batching.html)
        follow_batch = None
        exclude_keys = None
        return Batch.from_data_list(batch, follow_batch, exclude_keys)
    elif isinstance(elem, torch.Tensor):
        # pad with zero
        if not all(v.shape == elem.shape for v in batch):
            max_len = max(v.shape[0] for v in batch)
            batch = [
                torch.concat([x, torch.zeros(max_len - x.shape[0], *x.shape[1:])])
                for x in batch
            ]

        elem = batch[0]
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum([x.numel() for x in batch])  # assumes all have the same size
            # numel = max([x.numel() for x in batch]) * len(batch)    # total size of tensors
            storage = elem.storage()._new_shared(
                numel
            )  # Creates a new storage in shared memory with the same data type
            out = elem.new(storage).view(-1, *list(elem.size()))

        return torch.stack(batch, 0, out=out)
    elif (
        elem_type.__module__ == "numpy"
        and elem_type.__name__ != "str_"
        and elem_type.__name__ != "string_"
    ):
        if elem_type.__name__ == "ndarray" or elem_type.__name__ == "memmap":
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return default_collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int_classes):
        return torch.tensor(batch)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, container_abcs.Mapping):
        return {key: default_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, "_fields"):  # namedtuple
        return elem_type(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, container_abcs.Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if all(all(isinstance(t, tuple) for t in elem) for elem in it):
            return batch
        if not all(len(elem) == elem_size for elem in it):
            # if its a list of ints then append ints and return
            it = iter(batch)
            if all(all(isinstance(s, int_classes) for s in elem) for elem in it):
                return batch
            # if its a list of strings then append lists and return
            it = iter(batch)
            if all(
                all(isinstance(s, string_classes) for s in elem) for elem in it
            ) or all(
                all(
                    all(isinstance(s, string_classes) for s in t)
                    and isinstance(t, tuple)
                    for t in elem
                )
                for elem in it
            ):
                return batch
            raise RuntimeError("each element in list of batch should be of equal size")
        transposed = zip(*batch)
        return [default_collate(samples) for samples in transposed]

    raise TypeError(default_collate_err_msg_format.format(elem_type))


def ignore_None_collate(batch):
    """
    default_collate wrapper that creates batches only of not None values.
    Useful for cases when the dataset.__getitem__ can return None because of some
    exception and then we will want to exclude that sample from the batch.
    """
    batch = [x for x in batch if x is not None]
    if len(batch) == 0:
        return None
    return default_collate(batch)


def get_train_dataset_loader(args: Namespace, split: Optional[str] = "train"):
    """Given arg configuration, return appropriate torch.DataLoader
    for train data loader

    Args:
        args (Namespace): args
        split (str, optional): dataset split. Defaults to "train".

    Returns:
        train_data_loader: iterator that returns batches
    """
    train_data = get_object(args.dataset_name, "dataset")(args, split)

    if args.class_bal:
        if args.strategy == "ddp":
            sampler = DistributedWeightedSampler(
                train_data,
                weights=train_data.weights,
                replacement=True,
                rank=args.global_rank,
                num_replicas=args.world_size,
                drop_last=True,
            )
        else:
            sampler = data.sampler.WeightedRandomSampler(
                weights=train_data.weights,
                num_samples=len(train_data),
                replacement=True,
            )
    else:
        if args.strategy == "ddp":
            sampler = torch.utils.data.distributed.DistributedSampler(
                train_data,
                shuffle=True,
                rank=args.global_rank,
                num_replicas=args.world_size,
                drop_last=True,
            )
        else:
            sampler = data.sampler.RandomSampler(train_data)

    train_data_loader = data.DataLoader(
        train_data,
        num_workers=args.num_workers,
        sampler=sampler,
        pin_memory=True,
        batch_size=args.batch_size,
        collate_fn=ignore_None_collate,
        drop_last=True,
    )

    return train_data_loader


def get_eval_dataset_loader(
    args: Namespace, split: Literal["train", "dev", "test"], shuffle=False
):
    """_summary_

    Args:
        args (Namespace): args
        split (Literal[&quot;train&quot;, &quot;dev&quot;, &quot;test&quot;]): dataset split.
        shuffle (bool, optional): whether to shuffle dataset. Defaults to False.

    Returns:
        data_loader: iterator that returns batches
    """
    if shuffle:
        rprint("[bold] Eval dataloader shuffle = True")

    eval_data = get_object(args.dataset_name, "dataset")(args, split)

    if args.strategy == "ddp":
        sampler = torch.utils.data.distributed.DistributedSampler(
            eval_data,
            shuffle=shuffle,
            rank=args.global_rank,
            num_replicas=args.world_size,
        )
    else:
        sampler = (
            torch.utils.data.sampler.RandomSampler(eval_data)
            if shuffle
            else torch.utils.data.sampler.SequentialSampler(eval_data)
        )
    data_loader = torch.utils.data.DataLoader(
        eval_data,
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        collate_fn=ignore_None_collate,
        pin_memory=True,
        drop_last=False,
        sampler=sampler,
    )

    return data_loader


def is_dist_avail_and_initialized():
    if not torch.distributed.is_available():
        return False
    if not torch.distributed.is_initialized():
        return False
    return True


class GatherLayer(torch.autograd.Function):
    """
    Gather tensors from all workers with support for backward propagation:
    This implementation does not cut the gradients as torch.distributed.all_gather does.
    """

    @staticmethod
    def forward(ctx, x):
        output = [
            torch.zeros_like(x) for _ in range(torch.distributed.get_world_size())
        ]
        torch.distributed.all_gather(output, x)
        return tuple(output)

    @staticmethod
    def backward(ctx, *grads):
        all_gradients = torch.stack(grads)
        torch.distributed.all_reduce(all_gradients)
        return all_gradients[torch.distributed.get_rank()]


def all_gather_with_grad(tensors):
    """
    Performs all_gather operation on the provided tensors.
    Graph remains connected for backward grad computation.
    """
    # Queue the gathered tensors
    world_size = torch.distributed.get_world_size()
    # There is no need for reduction in the single-proc case
    if world_size == 1:
        return tensors

    # tensor_all = GatherLayer.apply(tensors)
    tensor_all = GatherLayer.apply(tensors)

    return torch.cat(tensor_all, dim=0)


@torch.no_grad()
def concat_all_gather(tensor):
    """
    Performs all_gather operation on the provided tensors.
    *** Warning ***: torch.distributed.all_gather has no gradient.
    """

    tensors_gather = [
        torch.ones_like(tensor) for _ in range(torch.distributed.get_world_size())
    ]
    torch.distributed.all_gather(tensors_gather, tensor, async_op=False)
    output = torch.cat(tensors_gather, dim=0)
    return output


def get_lightning_model(args: Namespace):
    """Create new model or load from checkpoint

    Args:
        args (Namespace): global args

    Raises:
        FileType: checkpoint_path must be ".args" or ".ckpt" file

    Returns:
        model: pl.LightningModule instance
    """
    if args.from_checkpoint:
        if args.checkpoint_path.endswith(".args"):
            snargs = Namespace(**pickle.load(open(args.checkpoint_path, "rb")))
            # update saved args with new arguments
            for k, v in vars(args).items():
                if k not in snargs:
                    setattr(snargs, k, v)
            model = get_object(snargs.lightning_name, "lightning")(snargs)
            modelpath = snargs.model_path
        elif args.checkpoint_path.endswith(".ckpt"):
            model = get_object(args.lightning_name, "lightning")(args)
            modelpath = args.checkpoint_path
            checkpoint = pl_load(
                args.checkpoint_path, map_location=lambda storage, loc: storage
            )
            snargs = checkpoint["hyper_parameters"]["args"]
        else:
            raise FileType("checkpoint_path should be an args or ckpt file.")
        # update args with old args if not found
        for k, v in vars(snargs).items():
            if k not in args:
                setattr(args, k, v)
        model = model.load_from_checkpoint(
            checkpoint_path=modelpath,
            strict=not args.relax_checkpoint_matching,
            **{"args": args},
        )
        rprint(f"[bold] Loaded checkpoint from {modelpath}")
    else:
        model = get_object(args.lightning_name, "lightning")(args)
    return model



import torch
from torch import optim
from clipzyme.utils.registry import register_object
from clipzyme.utils.classes import Nox


@register_object("reduce_on_plateau", "scheduler")
class ReduceLROnPlateau(optim.lr_scheduler.ReduceLROnPlateau, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.ReduceLROnPlateau.html#torch.optim.lr_scheduler.ReduceLROnPlateau
    """

    def __init__(self, optimizer, args):
        super().__init__(
            optimizer,
            patience=args.patience,
            factor=args.lr_decay,
            mode="min" if "loss" in args.monitor else "max",
        )


@register_object("exponential_decay", "scheduler")
class ExponentialLR(optim.lr_scheduler.ExponentialLR, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.ExponentialLR.html#torch.optim.lr_scheduler.ExponentialLR
    """

    def __init__(self, optimizer, args):
        super().__init__(optimizer, gamma=args.lr_decay)


@register_object("cosine_annealing", "scheduler")
class CosineAnnealingLR(optim.lr_scheduler.CosineAnnealingLR, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingLR.html
    """

    def __init__(self, optimizer, args):
        super().__init__(optimizer, args.cosine_annealing_period)


@register_object("cosine_annealing_restarts", "scheduler")
class CosineAnnealingWarmRestarts(optim.lr_scheduler.CosineAnnealingWarmRestarts, Nox):
    """
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingWarmRestarts.html#torch.optim.lr_scheduler.CosineAnnealingWarmRestarts
    """

    def __init__(self, optimizer, args):
        super().__init__(
            optimizer,
            T_0=args.cosine_annealing_period,
            T_mult=args.cosine_annealing_period_scaling,
        )



import hashlib

REGISTRIES = {
    "LIGHTNING_REGISTRY": {},
    "DATASET_REGISTRY": {},
    "MODEL_REGISTRY": {},
    "LOSS_REGISTRY": {},
    "METRIC_REGISTRY": {},
    "OPTIMIZER_REGISTRY": {},
    "SCHEDULER_REGISTRY": {},
    "SEARCHER_REGISTRY": {},
    "CALLBACK_REGISTRY": {},
    "INPUT_LOADER_REGISTRY": {},
    "AUGMENTATION_REGISTRY": {},
    "LOGGER_REGISTRY": {},
}


def get_object(object_name, object_type):
    if object_name not in REGISTRIES["{}_REGISTRY".format(object_type.upper())]:
        raise Exception(
            "INVALID {} NAME: {}. AVAILABLE {}".format(
                object_type.upper(),
                object_name,
                REGISTRIES["{}_REGISTRY".format(object_type.upper())].keys(),
            )
        )
    return REGISTRIES["{}_REGISTRY".format(object_type.upper())][object_name]


def register_object(object_name, object_type):
    def decorator(obj):
        REGISTRIES["{}_REGISTRY".format(object_type.upper())][object_name] = obj
        obj.name = object_name
        return obj

    return decorator


def md5(key):
    """
    returns a hashed with md5 string of the key
    """
    return hashlib.md5(key.encode()).hexdigest()
